r"""
Asymptotic Expansions --- Miscellaneous

AUTHORS:

- Daniel Krenn (2015)

ACKNOWLEDGEMENT:

- Benjamin Hackl, Clemens Heuberger and Daniel Krenn are supported by the
  Austrian Science Fund (FWF): P 24644-N26.

- Benjamin Hackl is supported by the Google Summer of Code 2015.


Functions, Classes and Methods
==============================
"""

# ****************************************************************************
# Copyright (C) 2015 Daniel Krenn <dev@danielkrenn.at>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import lazy_import
from sage.structure.sage_object import SageObject

lazy_import('sage.rings.lazy_series_ring', 'LazyPowerSeriesRing')
lazy_import('sage.rings.multi_power_series_ring', 'MPowerSeriesRing_generic')
lazy_import('sage.rings.polynomial.multi_polynomial_ring_base', 'MPolynomialRing_base')
lazy_import('sage.rings.polynomial.polynomial_ring', 'PolynomialRing_generic')
lazy_import('sage.rings.power_series_ring', 'PowerSeriesRing_generic')


def repr_short_to_parent(s):
    r"""
    Helper method for the growth group factory, which converts a short
    representation string to a parent.

    INPUT:

    - ``s`` -- string; short representation of a parent

    OUTPUT: a parent

    The possible short representations are shown in the examples below.

    EXAMPLES::

        sage: from sage.rings.asymptotic.misc import repr_short_to_parent
        sage: repr_short_to_parent('ZZ')
        Integer Ring
        sage: repr_short_to_parent('QQ')
        Rational Field
        sage: repr_short_to_parent('SR')
        Symbolic Ring
        sage: repr_short_to_parent('NN')
        Non negative integer semiring
        sage: repr_short_to_parent('UU')
        Group of Roots of Unity

    TESTS::

        sage: repr_short_to_parent('abcdef')
        Traceback (most recent call last):
        ...
        ValueError: Cannot create a parent out of 'abcdef'.
        > *previous* ValueError: unknown specification abcdef
        > *and* NameError: name 'abcdef' is not defined
    """
    from sage.groups.misc_gps.argument_groups import ArgumentGroup
    from sage.misc.sage_eval import sage_eval

    def extract(s):
        try:
            return ArgumentGroup(specification=s)
        except Exception as e:
            e_ag = e
            e_ag.__traceback__ = None

        try:
            return sage_eval(s)
        except Exception as e:
            e_se = e
            e_se.__traceback__ = None

        raise combine_exceptions(
            ValueError("Cannot create a parent out of '%s'." % (s,)),
            e_ag, e_se)

    P = extract(s)

    from sage.misc.lazy_import import LazyImport
    if type(P) is LazyImport:
        P = P._get_object()

    from sage.structure.parent import Parent
    if not isinstance(P, Parent):
        raise ValueError("'%s' does not describe a parent." % (s,))
    return P


def parent_to_repr_short(P):
    r"""
    Helper method which generates a short(er) representation string
    out of a parent.

    INPUT:

    - ``P`` -- a parent

    OUTPUT: string

    EXAMPLES::

        sage: from sage.rings.asymptotic.misc import parent_to_repr_short
        sage: parent_to_repr_short(ZZ)
        'ZZ'
        sage: parent_to_repr_short(QQ)
        'QQ'
        sage: parent_to_repr_short(SR)
        'SR'
        sage: parent_to_repr_short(RR)
        'RR'
        sage: parent_to_repr_short(CC)
        'CC'
        sage: parent_to_repr_short(ZZ['x'])
        'ZZ[x]'
        sage: parent_to_repr_short(QQ['d, k'])
        'QQ[d, k]'
        sage: parent_to_repr_short(QQ['e'])
        'QQ[e]'
        sage: parent_to_repr_short(SR[['a, r']])
        'SR[[a, r]]'
        sage: parent_to_repr_short(Zmod(3))
        'Ring of integers modulo 3'
        sage: parent_to_repr_short(Zmod(3)['g'])
        'Univariate Polynomial Ring in g over Ring of integers modulo 3'
    """
    from sage.rings.cc import CC
    from sage.rings.cif import CIF
    from sage.rings.complex_arb import CBF
    from sage.rings.integer_ring import ZZ
    from sage.rings.rational_field import QQ
    from sage.rings.real_arb import RBF
    from sage.rings.real_mpfi import RIF
    from sage.rings.real_mpfr import RR
    from sage.symbolic.ring import SR

    def abbreviate(P):
        try:
            return P._repr_short_()
        except AttributeError:
            pass
        abbreviations = {ZZ: 'ZZ', QQ: 'QQ', SR: 'SR',
                         RR: 'RR', CC: 'CC',
                         RIF: 'RIF', CIF: 'CIF',
                         RBF: 'RBF', CBF: 'CBF'}
        try:
            return abbreviations[P]
        except KeyError:
            pass
        raise ValueError('Cannot abbreviate %s.' % (P,))

    poly = isinstance(P, (PolynomialRing_generic, MPolynomialRing_base))
    power = isinstance(P, (PowerSeriesRing_generic, MPowerSeriesRing_generic, LazyPowerSeriesRing))

    if poly or power:
        if poly:
            op, cl = ('[', ']')
        else:
            op, cl = ('[[', ']]')
        try:
            s = abbreviate(P.base_ring()) + op + ', '.join(P.variable_names()) + cl
        except ValueError:
            s = str(P)
    else:
        try:
            s = abbreviate(P)
        except ValueError:
            s = str(P)

    return s


def split_str_by_op(string, op, strip_parentheses=True):
    r"""
    Split the given string into a tuple of substrings arising by
    splitting by ``op`` and taking care of parentheses.

    INPUT:

    - ``string`` -- string

    - ``op`` -- string; this is used by
      :python:`str.split <library/stdtypes.html#str.split>`.
      Thus, if this is ``None``, then any whitespace string is a
      separator and empty strings are removed from the result.

    - ``strip_parentheses`` -- boolean (default: ``True``)

    OUTPUT: a tuple of strings

    TESTS::

        sage: from sage.rings.asymptotic.misc import split_str_by_op
        sage: split_str_by_op('x^ZZ', '*')
        ('x^ZZ',)
        sage: split_str_by_op('log(x)^ZZ * y^QQ', '*')
        ('log(x)^ZZ', 'y^QQ')
        sage: split_str_by_op('log(x)**ZZ * y**QQ', '*')
        ('log(x)**ZZ', 'y**QQ')
        sage: split_str_by_op('a^b * * c^d', '*')
        Traceback (most recent call last):
        ...
        ValueError: 'a^b * * c^d' is invalid since a '*' follows a '*'.
        sage: split_str_by_op('a^b * (c*d^e)', '*')
        ('a^b', 'c*d^e')

    ::

        sage: split_str_by_op('(a^b)^c', '^')
        ('a^b', 'c')
        sage: split_str_by_op('a^(b^c)', '^')
        ('a', 'b^c')

    ::

        sage: split_str_by_op('(a) + (b)', op='+', strip_parentheses=True)
        ('a', 'b')
        sage: split_str_by_op('(a) + (b)', op='+', strip_parentheses=False)
        ('(a)', '(b)')
        sage: split_str_by_op(' ( t  ) ', op='+', strip_parentheses=False)
        ('( t  )',)

    ::

        sage: split_str_by_op(' ( t  ) ', op=None)
        ('t',)
        sage: split_str_by_op(' ( t  )s', op=None)
        ('(t)s',)
        sage: split_str_by_op(' ( t  ) s', op=None)
        ('t', 's')

    ::

        sage: split_str_by_op('(e^(n*log(n)))^SR.subring(no_variables=True)', '*')
        ('(e^(n*log(n)))^SR.subring(no_variables=True)',)
    """
    def is_balanced(s):
        open = 0
        for l in s:
            if l == '(':
                open += 1
            elif l == ')':
                open -= 1
            if open < 0:
                return False
        return bool(open == 0)

    factors = []
    balanced = True
    if string and op is not None and string.startswith(op):
        raise ValueError("'%s' is invalid since it starts with a '%s'." %
                         (string, op))
    for s in string.split(op):
        if not s:
            factors[-1] += op
            balanced = False
            continue
        if not s.strip():
            raise ValueError("'%s' is invalid since a '%s' follows a '%s'." %
                             (string, op, op))
        if not balanced:
            s = factors.pop() + (op if op else '') + s
        balanced = is_balanced(s)
        factors.append(s)

    if not balanced:
        raise ValueError("Parentheses in '%s' are not balanced." % (string,))

    def strip(s):
        s = s.strip()
        if not s:
            return s
        if strip_parentheses and s[0] == '(' and s[-1] == ')':
            t = s[1:-1]
            if is_balanced(t):
                s = t
        return s.strip()

    return tuple(strip(f) for f in factors)


def repr_op(left, op, right=None, latex=False):
    r"""
    Create a string ``left op right`` with
    taking care of parentheses in its operands.

    INPUT:

    - ``left`` -- an element

    - ``op`` -- string

    - ``right`` -- an element

    - ``latex`` -- boolean (default: ``False``); if set, then
      LaTeX-output is returned

    OUTPUT: string

    EXAMPLES::

        sage: from sage.rings.asymptotic.misc import repr_op
        sage: repr_op('a^b', '^', 'c')
        '(a^b)^c'

    TESTS::

        sage: repr_op('a-b', '^', 'c')
        '(a-b)^c'
        sage: repr_op('a+b', '^', 'c')
        '(a+b)^c'

    ::

        sage: print(repr_op(r'\frac{1}{2}', '^', 'c', latex=True))
        \left(\frac{1}{2}\right)^c

    ::

        sage: repr_op('Arg', '_', 'Symbolic Ring')
        'Arg_(Symbolic Ring)'
    """
    left = str(left)
    right = str(right) if right is not None else ''

    def add_parentheses(s, op):
        if op in ('^', '_'):
            signals = ('^', '/', '*', '-', '+', ' ')
        else:
            return s
        if any(sig in s for sig in signals) or latex and s.startswith(r'\frac'):
            if latex:
                return r'\left({}\right)'.format(s)
            return '({})'.format(s)
        return s

    return add_parentheses(left, op) + op + add_parentheses(right, op)


def combine_exceptions(e, *f):
    r"""
    Helper function which combines the messages of the given exceptions.

    INPUT:

    - ``e`` -- an exception

    - ``*f`` -- exceptions

    OUTPUT: an exception

    EXAMPLES::

        sage: from sage.rings.asymptotic.misc import combine_exceptions
        sage: raise combine_exceptions(ValueError('Outer.'), TypeError('Inner.'))
        Traceback (most recent call last):
        ...
        ValueError: Outer.
        > *previous* TypeError: Inner.
        sage: raise combine_exceptions(ValueError('Outer.'),
        ....:                          TypeError('Inner1.'), TypeError('Inner2.'))
        Traceback (most recent call last):
        ...
        ValueError: Outer.
        > *previous* TypeError: Inner1.
        > *and* TypeError: Inner2.
        sage: raise combine_exceptions(ValueError('Outer.'),
        ....:                          combine_exceptions(TypeError('Middle.'),
        ....:                                             TypeError('Inner.')))
        Traceback (most recent call last):
        ...
        ValueError: Outer.
        > *previous* TypeError: Middle.
        >> *previous* TypeError: Inner.
    """
    import re
    msg = ('\n *previous* ' +
           '\n *and* '.join("%s: %s" % (ff.__class__.__name__, str(ff)) for ff in f))
    msg = re.sub(r'^([>]* \*previous\*)', r'>\1', msg, flags=re.MULTILINE)
    msg = re.sub(r'^([>]* \*and\*)', r'>\1', msg, flags=re.MULTILINE)
    msg = str(e.args if len(e.args) > 1 else e.args[0]) + msg
    e.args = (msg,)
    return e


def substitute_raise_exception(element, e):
    r"""
    Raise an error describing what went wrong with the substitution.

    INPUT:

    - ``element`` -- an element

    - ``e`` -- an exception which is included in the raised error message

    OUTPUT: raise an exception of the same type as ``e``

    TESTS::

        sage: from sage.rings.asymptotic.misc import substitute_raise_exception
        sage: substitute_raise_exception(x, Exception('blub'))
        Traceback (most recent call last):
        ...
        Exception: Cannot substitute in x in Symbolic Ring.
        > *previous* Exception: blub
    """
    raise combine_exceptions(
        type(e)('Cannot substitute in %s in %s.' %
                (element, element.parent())), e)


def bidirectional_merge_overlapping(A, B, key=None):
    r"""
    Merge the two overlapping tuples/lists.

    INPUT:

    - ``A`` -- list or tuple (type has to coincide with type of ``B``)

    - ``B`` -- list or tuple (type has to coincide with type of ``A``)

    - ``key`` -- (default: ``None``) a function. If ``None``, then the
      identity is used.  This ``key``-function applied on an element
      of the list/tuple is used for comparison. Thus elements with the
      same key are considered as equal.

    OUTPUT:

    A pair of lists or tuples (depending on the type of ``A`` and ``B``).

    .. NOTE::

        Suppose we can decompose the list `A=ac` and `B=cb` with
        lists `a`, `b`, `c`, where `c` is nonempty. Then
        :func:`bidirectional_merge_overlapping` returns the pair `(acb, acb)`.

        Suppose a ``key``-function is specified and `A=ac_A` and
        `B=c_Bb`, where the list of keys of the elements of `c_A`
        equals the list of keys of the elements of `c_B`. Then
        :func:`bidirectional_merge_overlapping` returns the pair `(ac_Ab, ac_Bb)`.

        After unsuccessfully merging `A=ac` and `B=cb`,
        a merge of `A=ca` and `B=bc` is tried.

    TESTS::

        sage: from sage.rings.asymptotic.misc import bidirectional_merge_overlapping
        sage: def f(L, s):
        ....:     return list((ell, s) for ell in L)
        sage: key = lambda k: k[0]
        sage: bidirectional_merge_overlapping(f([0..3], 'a'), f([5..7], 'b'), key)
        Traceback (most recent call last):
        ...
        ValueError: Input does not have an overlap.
        sage: bidirectional_merge_overlapping(f([0..2], 'a'), f([4..7], 'b'), key)
        Traceback (most recent call last):
        ...
        ValueError: Input does not have an overlap.
        sage: bidirectional_merge_overlapping(f([4..7], 'a'), f([0..2], 'b'), key)
        Traceback (most recent call last):
        ...
        ValueError: Input does not have an overlap.
        sage: bidirectional_merge_overlapping(f([0..3], 'a'), f([3..4], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a'), (4, 'b')],
         [(0, 'a'), (1, 'a'), (2, 'a'), (3, 'b'), (4, 'b')])
        sage: bidirectional_merge_overlapping(f([3..4], 'a'), f([0..3], 'b'), key)
        ([(0, 'b'), (1, 'b'), (2, 'b'), (3, 'a'), (4, 'a')],
         [(0, 'b'), (1, 'b'), (2, 'b'), (3, 'b'), (4, 'a')])
        sage: bidirectional_merge_overlapping(f([0..1], 'a'), f([0..4], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'b'), (3, 'b'), (4, 'b')],
         [(0, 'b'), (1, 'b'), (2, 'b'), (3, 'b'), (4, 'b')])
        sage: bidirectional_merge_overlapping(f([0..3], 'a'), f([0..1], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a')],
         [(0, 'b'), (1, 'b'), (2, 'a'), (3, 'a')])
        sage: bidirectional_merge_overlapping(f([0..3], 'a'), f([1..3], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a')],
         [(0, 'a'), (1, 'b'), (2, 'b'), (3, 'b')])
        sage: bidirectional_merge_overlapping(f([1..3], 'a'), f([0..3], 'b'), key)
        ([(0, 'b'), (1, 'a'), (2, 'a'), (3, 'a')],
         [(0, 'b'), (1, 'b'), (2, 'b'), (3, 'b')])
        sage: bidirectional_merge_overlapping(f([0..6], 'a'), f([3..4], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a'), (4, 'a'), (5, 'a'), (6, 'a')],
         [(0, 'a'), (1, 'a'), (2, 'a'), (3, 'b'), (4, 'b'), (5, 'a'), (6, 'a')])
        sage: bidirectional_merge_overlapping(f([0..3], 'a'), f([1..2], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a')],
         [(0, 'a'), (1, 'b'), (2, 'b'), (3, 'a')])
        sage: bidirectional_merge_overlapping(f([1..2], 'a'), f([0..3], 'b'), key)
        ([(0, 'b'), (1, 'a'), (2, 'a'), (3, 'b')],
         [(0, 'b'), (1, 'b'), (2, 'b'), (3, 'b')])
        sage: bidirectional_merge_overlapping(f([1..3], 'a'), f([1..3], 'b'), key)
        ([(1, 'a'), (2, 'a'), (3, 'a')],
         [(1, 'b'), (2, 'b'), (3, 'b')])
    """
    if key is None:
        Akeys = A
        Bkeys = B
    else:
        Akeys = tuple(key(a) for a in A)
        Bkeys = tuple(key(b) for b in B)

    def find_overlapping_index(A, B):
        if len(B) > len(A) - 2:
            raise StopIteration
        matches = iter(i for i in range(1, len(A) - len(B))
                       if A[i:i+len(B)] == B)
        return next(matches)

    def find_mergedoverlapping_index(A, B):
        """
        Return in index i where to merge two overlapping tuples/lists ``A`` and ``B``.

        Then ``A + B[i:]`` or ``A[:-i] + B`` are the merged tuples/lists.

        Adapted from https://stackoverflow.com/a/30056066/1052778.
        """
        matches = iter(i for i in range(min(len(A), len(B)), 0, -1)
                       if A[-i:] == B[:i])
        return next(matches, 0)

    i = find_mergedoverlapping_index(Akeys, Bkeys)
    if i > 0:
        return A + B[i:], A[:-i] + B

    i = find_mergedoverlapping_index(Bkeys, Akeys)
    if i > 0:
        return B[:-i] + A, B + A[i:]

    try:
        i = find_overlapping_index(Akeys, Bkeys)
    except StopIteration:
        pass
    else:
        return A, A[:i] + B + A[i+len(B):]

    try:
        i = find_overlapping_index(Bkeys, Akeys)
    except StopIteration:
        pass
    else:
        return B[:i] + A + B[i+len(A):], B

    raise ValueError('Input does not have an overlap.')


def bidirectional_merge_sorted(A, B, key=None):
    r"""
    Merge the two tuples/lists, keeping the orders provided by them.

    INPUT:

    - ``A`` -- list or tuple (type has to coincide with type of ``B``)

    - ``B`` -- list or tuple (type has to coincide with type of ``A``)

    - ``key`` -- (default: ``None``) a function. If ``None``, then the
      identity is used.  This ``key``-function applied on an element
      of the list/tuple is used for comparison. Thus elements with the
      same key are considered as equal.

    .. NOTE::

        The two tuples/list need to overlap, i.e. need at least
        one key in common.

    OUTPUT:

    A pair of lists containing all elements totally ordered. (The first
    component uses ``A`` as a merge base, the second component ``B``.)

    If merging fails, then a
    :python:`RuntimeError<library/exceptions.html#exceptions.RuntimeError>`
    is raised.

    TESTS::

        sage: from sage.rings.asymptotic.misc import bidirectional_merge_sorted
        sage: def f(L, s):
        ....:     return list((ell, s) for ell in L)
        sage: key = lambda k: k[0]
        sage: bidirectional_merge_sorted(f([0..3], 'a'), f([5..7], 'b'), key)
        Traceback (most recent call last):
        ...
        RuntimeError: no common elements
        sage: bidirectional_merge_sorted(f([0..2], 'a'), f([4..7], 'b'), key)
        Traceback (most recent call last):
        ...
        RuntimeError: no common elements
        sage: bidirectional_merge_sorted(f([4..7], 'a'), f([0..2], 'b'), key)
        Traceback (most recent call last):
        ...
        RuntimeError: no common elements
        sage: bidirectional_merge_sorted(f([0..3], 'a'), f([3..4], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a'), (4, 'b')],
         [(0, 'a'), (1, 'a'), (2, 'a'), (3, 'b'), (4, 'b')])
        sage: bidirectional_merge_sorted(f([3..4], 'a'), f([0..3], 'b'), key)
        ([(0, 'b'), (1, 'b'), (2, 'b'), (3, 'a'), (4, 'a')],
         [(0, 'b'), (1, 'b'), (2, 'b'), (3, 'b'), (4, 'a')])
        sage: bidirectional_merge_sorted(f([0..1], 'a'), f([0..4], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'b'), (3, 'b'), (4, 'b')],
         [(0, 'b'), (1, 'b'), (2, 'b'), (3, 'b'), (4, 'b')])
        sage: bidirectional_merge_sorted(f([0..3], 'a'), f([0..1], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a')],
         [(0, 'b'), (1, 'b'), (2, 'a'), (3, 'a')])
        sage: bidirectional_merge_sorted(f([0..3], 'a'), f([1..3], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a')],
         [(0, 'a'), (1, 'b'), (2, 'b'), (3, 'b')])
        sage: bidirectional_merge_sorted(f([1..3], 'a'), f([0..3], 'b'), key)
        ([(0, 'b'), (1, 'a'), (2, 'a'), (3, 'a')],
         [(0, 'b'), (1, 'b'), (2, 'b'), (3, 'b')])
        sage: bidirectional_merge_sorted(f([0..6], 'a'), f([3..4], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a'), (4, 'a'), (5, 'a'), (6, 'a')],
         [(0, 'a'), (1, 'a'), (2, 'a'), (3, 'b'), (4, 'b'), (5, 'a'), (6, 'a')])
        sage: bidirectional_merge_sorted(f([0..3], 'a'), f([1..2], 'b'), key)
        ([(0, 'a'), (1, 'a'), (2, 'a'), (3, 'a')],
         [(0, 'a'), (1, 'b'), (2, 'b'), (3, 'a')])
        sage: bidirectional_merge_sorted(f([1..2], 'a'), f([0..3], 'b'), key)
        ([(0, 'b'), (1, 'a'), (2, 'a'), (3, 'b')],
         [(0, 'b'), (1, 'b'), (2, 'b'), (3, 'b')])
        sage: bidirectional_merge_sorted(f([1..3], 'a'), f([1..3], 'b'), key)
        ([(1, 'a'), (2, 'a'), (3, 'a')],
         [(1, 'b'), (2, 'b'), (3, 'b')])

    ::

        sage: bidirectional_merge_sorted(f([1, 2, 3], 'a'), f([1, 3], 'b'), key)
        ([(1, 'a'), (2, 'a'), (3, 'a')],
         [(1, 'b'), (2, 'a'), (3, 'b')])
        sage: bidirectional_merge_sorted(f([1, 4, 5, 6], 'a'), f([1, 2, 3, 4, 6], 'b'), key)
        ([(1, 'a'), (2, 'b'), (3, 'b'), (4, 'a'), (5, 'a'), (6, 'a')],
         [(1, 'b'), (2, 'b'), (3, 'b'), (4, 'b'), (5, 'a'), (6, 'b')])
        sage: bidirectional_merge_sorted(f([1, 2, 3, 4], 'a'), f([1, 3, 5], 'b'), key)
        Traceback (most recent call last):
        ...
        RuntimeError: sorting not unique
        sage: bidirectional_merge_sorted(f([1, 2], 'a'), f([2, 1], 'b'), key)
        Traceback (most recent call last):
        ...
        RuntimeError: sorting in lists not compatible
        sage: bidirectional_merge_sorted(f([1, 2, 4], 'a'), f([1, 3, 4], 'b'), key)
        Traceback (most recent call last):
        ...
        RuntimeError: sorting not unique
    """
    if key is None:
        Akeys = A
        Bkeys = B
    else:
        Akeys = tuple(key(a) for a in A)
        Bkeys = tuple(key(b) for b in B)

    matches = tuple((i, j)
                    for i, a in enumerate(Akeys)
                    for j, b in enumerate(Bkeys)
                    if a == b)
    if not matches:
        raise RuntimeError('no common elements')

    resultA = []
    resultB = []
    last = (0, 0)
    end = (len(A), len(B))
    for current in matches + (end,):
        if not all(a <= b for a, b in zip(last, current)):
            raise RuntimeError('sorting in lists not compatible')
        if last[0] == current[0]:
            resultA.extend(B[last[1]:current[1]])
            resultB.extend(B[last[1]:current[1]])
        elif last[1] == current[1]:
            resultA.extend(A[last[0]:current[0]])
            resultB.extend(A[last[0]:current[0]])
        else:
            raise RuntimeError('sorting not unique')
        if current != end:
            resultA.append(A[current[0]])
            resultB.append(B[current[1]])
            last = (current[0]+1, current[1]+1)

    return (resultA, resultB)


def log_string(element, base=None):
    r"""
    Return a representation of the log of the given element to the
    given base.

    INPUT:

    - ``element`` -- an object

    - ``base`` -- an object or ``None``

    OUTPUT: string

    EXAMPLES::

        sage: from sage.rings.asymptotic.misc import log_string
        sage: log_string(3)
        'log(3)'
        sage: log_string(3, base=42)
        'log(3, base=42)'
    """
    basestr = ', base=' + str(base) if base else ''
    return 'log(%s%s)' % (element, basestr)


def strip_symbolic(expression):
    r"""
    Return, if possible, the underlying (numeric) object of
    the symbolic expression.

    If ``expression`` is not symbolic, then ``expression`` is returned.

    INPUT:

    - ``expression`` -- an object

    OUTPUT: an object

    EXAMPLES::

        sage: from sage.rings.asymptotic.misc import strip_symbolic
        sage: strip_symbolic(SR(2)); _.parent()
        2
        Integer Ring
        sage: strip_symbolic(SR(2/3)); _.parent()
        2/3
        Rational Field
        sage: strip_symbolic(SR('x')); _.parent()
        x
        Symbolic Ring
        sage: strip_symbolic(pi); _.parent()
        pi
        Symbolic Ring
    """
    from sage.structure.element import parent, Element
    from sage.symbolic.ring import SymbolicRing

    P = parent(expression)
    if isinstance(P, SymbolicRing):
        try:
            stripped = expression.pyobject()
            if isinstance(stripped, Element):
                return stripped
        except TypeError:
            pass
    return expression


class NotImplementedOZero(NotImplementedError):
    r"""
    A special :python:`NotImplementedError<library/exceptions.html#exceptions.NotImplementedError>`
    which is raised when the result is O(0) which means 0
    for sufficiently large values of the variable.
    """
    def __init__(self, asymptotic_ring=None, var=None, exact_part=0):
        r"""
        INPUT:

        - ``asymptotic_ring`` -- (default: ``None``) an :class:`AsymptoticRing`
          or ``None``

        - ``var`` -- (default: ``None``) a string

        Either ``asymptotic_ring`` or ``var`` has to be specified.

        - ``exact_part`` -- (default: ``0``) asymptotic expansion

        EXAMPLES::

            sage: A = AsymptoticRing('n^ZZ', ZZ)
            sage: from sage.rings.asymptotic.misc import NotImplementedOZero

            sage: raise NotImplementedOZero(A)
            Traceback (most recent call last):
            ...
            NotImplementedOZero: got O(0)
            The error term O(0) means 0 for sufficiently large n.

            sage: raise NotImplementedOZero(var='m')
            Traceback (most recent call last):
            ...
            NotImplementedOZero: got O(0)
            The error term O(0) means 0 for sufficiently large m.

        TESTS::

            sage: raise NotImplementedOZero(A, var='m')
            Traceback (most recent call last):
            ...
            ValueError: specify either 'asymptotic_ring' or 'var'
            sage: raise NotImplementedOZero()
            Traceback (most recent call last):
            ...
            ValueError: specify either 'asymptotic_ring' or 'var'
        """
        if (asymptotic_ring is None) == (var is None):
            raise ValueError("specify either 'asymptotic_ring' or 'var'")

        if var is None:
            var = ', '.join(str(g) for g in asymptotic_ring.gens())
        message = ('got {}\n'.format(('{} + '.format(exact_part) if exact_part else '')
                                     + 'O(0)') +
                   'The error term O(0) '
                   'means 0 for sufficiently large {}.'.format(var))

        if asymptotic_ring is not None and isinstance(exact_part, int) and exact_part == 0:
            exact_part = asymptotic_ring.zero()
        self.exact_part = exact_part

        super().__init__(message)


class NotImplementedBZero(NotImplementedError):
    r"""
    A special :python:`NotImplementedError<library/exceptions.html#exceptions.NotImplementedError>`
    which is raised when the result is B(0) which means 0
    for sufficiently large values of the variable.
    """
    def __init__(self, asymptotic_ring=None, var=None, exact_part=0):
        r"""
        INPUT:

        - ``asymptotic_ring`` -- (default: ``None``) an :class:`AsymptoticRing`
          or ``None``

        - ``var`` -- (default: ``None``) string

        Either ``asymptotic_ring`` or ``var`` has to be specified.

        - ``exact_part`` -- (default: ``0``) asymptotic expansion

        EXAMPLES::

            sage: A = AsymptoticRing('n^ZZ', ZZ)
            sage: from sage.rings.asymptotic.misc import NotImplementedBZero

            sage: raise NotImplementedBZero(A)
            Traceback (most recent call last):
            ...
            NotImplementedBZero: got B(0)
            The error term B(0) means 0 for sufficiently large n.

            sage: raise NotImplementedBZero(var='m')
            Traceback (most recent call last):
            ...
            NotImplementedBZero: got B(0)
            The error term B(0) means 0 for sufficiently large m.

            sage: AR.<n> = AsymptoticRing('n^QQ', QQ)
            sage: AR(0).B(42)
            Traceback (most recent call last):
            ...
            NotImplementedBZero: got B(0)
            The error term B(0) means 0 for sufficiently large n.

        TESTS::

            sage: raise NotImplementedBZero(A, var='m')
            Traceback (most recent call last):
            ...
            ValueError: specify either 'asymptotic_ring' or 'var'
            sage: raise NotImplementedBZero()
            Traceback (most recent call last):
            ...
            ValueError: specify either 'asymptotic_ring' or 'var'
        """
        if (asymptotic_ring is None) == (var is None):
            raise ValueError("specify either 'asymptotic_ring' or 'var'")

        if var is None:
            var = ', '.join(str(g) for g in asymptotic_ring.gens())
        message = ('got {}\n'.format(('{} + '.format(exact_part) if exact_part else '')
                                     + 'B(0)') +
                   'The error term B(0) '
                   'means 0 for sufficiently large {}.'.format(var))

        if asymptotic_ring is not None and isinstance(exact_part, int) and exact_part == 0:
            exact_part = asymptotic_ring.zero()
        self.exact_part = exact_part

        super().__init__(message)


def transform_category(category,
                       subcategory_mapping, axiom_mapping,
                       initial_category=None):
    r"""
    Transform ``category`` to a new category according to the given
    mappings.

    INPUT:

    - ``category`` -- a category

    - ``subcategory_mapping`` -- list (or other iterable) of triples
      ``(from, to, mandatory)``, where

      - ``from`` and ``to`` are categories and
      - ``mandatory`` is a boolean.

    - ``axiom_mapping`` -- list (or other iterable) of triples
      ``(from, to, mandatory)``, where

      - ``from`` and ``to`` are strings describing axioms and
      - ``mandatory`` is a boolean.

    - ``initial_category`` -- (default: ``None``) a category. When
      transforming the given category, this ``initial_category`` is
      used as a starting point of the result. This means the resulting
      category will be a subcategory of ``initial_category``.
      If ``initial_category`` is ``None``, then the
      :class:`category of objects <sage.categories.objects.Objects>`
      is used.

    OUTPUT: a category

    .. NOTE::

        Consider a subcategory mapping ``(from, to, mandatory)``. If
        ``category`` is a subcategory of ``from``, then the
        returned category will be a subcategory of ``to``. Otherwise and
        if ``mandatory`` is set, then an error is raised.

        Consider an axiom mapping ``(from, to, mandatory)``. If
        ``category`` is has axiom ``from``, then the
        returned category will have axiom ``to``. Otherwise and
        if ``mandatory`` is set, then an error is raised.

    EXAMPLES::

        sage: from sage.rings.asymptotic.misc import transform_category
        sage: from sage.categories.additive_semigroups import AdditiveSemigroups
        sage: from sage.categories.additive_monoids import AdditiveMonoids
        sage: from sage.categories.additive_groups import AdditiveGroups
        sage: S = [
        ....:     (Sets(), Sets(), True),
        ....:     (Posets(), Posets(), False),
        ....:     (AdditiveMagmas(), Magmas(), False)]
        sage: A = [
        ....:     ('AdditiveAssociative', 'Associative', False),
        ....:     ('AdditiveUnital', 'Unital', False),
        ....:     ('AdditiveInverse', 'Inverse', False),
        ....:     ('AdditiveCommutative', 'Commutative', False)]
        sage: transform_category(Objects(), S, A)
        Traceback (most recent call last):
        ...
        ValueError: Category of objects is not
        a subcategory of Category of sets.
        sage: transform_category(Sets(), S, A)
        Category of sets
        sage: transform_category(Posets(), S, A)
        Category of posets
        sage: transform_category(AdditiveSemigroups(), S, A)
        Category of semigroups
        sage: transform_category(AdditiveMonoids(), S, A)
        Category of monoids
        sage: transform_category(AdditiveGroups(), S, A)
        Category of groups
        sage: transform_category(AdditiveGroups().AdditiveCommutative(), S, A)
        Category of commutative groups

    ::

        sage: transform_category(AdditiveGroups().AdditiveCommutative(), S, A,
        ....:     initial_category=Posets())
        Join of Category of commutative groups
            and Category of posets

    ::

        sage: transform_category(ZZ.category(), S, A)
        Category of commutative groups
        sage: transform_category(QQ.category(), S, A)
        Category of commutative groups
        sage: transform_category(SR.category(), S, A)
        Category of commutative groups
        sage: transform_category(Fields(), S, A)
        Category of commutative groups
        sage: transform_category(ZZ['t'].category(), S, A)
        Category of commutative groups

    ::

        sage: A[-1] = ('Commutative', 'AdditiveCommutative', True)
        sage: transform_category(Groups(), S, A)
        Traceback (most recent call last):
        ...
        ValueError: Category of groups does not have
        axiom Commutative.
    """
    if initial_category is None:
        from sage.categories.objects import Objects
        result = Objects()
    else:
        result = initial_category

    for A, B, mandatory in subcategory_mapping:
        if category.is_subcategory(A):
            result &= B
        elif mandatory:
            raise ValueError('%s is not a subcategory of %s.' %
                             (category, A))

    axioms = category.axioms()
    for A, B, mandatory in axiom_mapping:
        if A in axioms:
            result = result._with_axiom(B)
        elif mandatory:
            raise ValueError('%s does not have axiom %s.' %
                             (category, A))

    return result


class Locals(dict):
    r"""
    A frozen dictionary-like class for storing locals
    of an :class:`~sage.rings.asymptotic.asymptotic_ring.AsymptoticRing`.

    EXAMPLES::

        sage: from sage.rings.asymptotic.misc import Locals
        sage: locals = Locals({'a': 42})
        sage: locals['a']
        42

    The object contains default values (see :meth:`default_locals`)
    for some keys::

        sage: locals['log']
        <function log at 0x...>
    """
    def __getitem__(self, key):
        r"""
        Return an item.

        TESTS::

            sage: from sage.rings.asymptotic.misc import Locals
            sage: locals = Locals()
            sage: locals
            {}
            sage: locals['log']  # indirect doctest
            <function log at 0x...>
        """
        try:
            return super().__getitem__(key)
        except KeyError as ke:
            try:
                return self.default_locals()[key]
            except KeyError:
                raise ke

    def __setitem__(self, key, value):
        r"""
        Set an item.

        This raises an error as the object is immutable.

        TESTS::

            sage: from sage.rings.asymptotic.misc import Locals
            sage: locals = Locals()
            sage: locals['a'] = 4  # indirect doctest
            Traceback (most recent call last):
            ...
            TypeError: locals dictionary is frozen,
            therefore does not support item assignment
        """
        raise TypeError('locals dictionary is frozen, therefore does not support item assignment')

    @cached_method
    def _data_(self):
        r"""
        Return stored data as tuple sorted by their keys.

        TESTS::

            sage: from sage.rings.asymptotic.misc import Locals
            sage: locals = Locals({'a': 2, 'b': 1})
            sage: locals._data_()
            (('a', 2), ('b', 1))
        """
        return tuple(sorted(self.items(), key=lambda k_v: k_v[0]))

    def __hash__(self):
        r"""
        Return a hash value.

        TESTS::

            sage: from sage.rings.asymptotic.misc import Locals
            sage: locals = Locals({'a': 2, 'b': 1})
            sage: hash(locals)  # random
            -42
        """
        return hash(self._data_())

    @cached_method
    def default_locals(self):
        r"""
        Return the default locals used in
        the :class:`~sage.rings.asymptotic.asymptotic_ring.AsymptoticRing`.

        OUTPUT: a dictionary

        EXAMPLES::

            sage: from sage.rings.asymptotic.misc import Locals
            sage: locals = Locals({'a': 2, 'b': 1})
            sage: locals
            {'a': 2, 'b': 1}
            sage: locals.default_locals()
            {'log': <function log at 0x...>}
            sage: locals['log']
            <function log at 0x...>
        """
        from sage.functions.log import log
        return {
            'log': log}


class WithLocals(SageObject):
    r"""
    A class extensions for handling local values; see also
    :class:`Locals`.

    This is used in the
    :class:`~sage.rings.asymptotic.asymptotic_ring.AsymptoticRing`.

    EXAMPLES::

        sage: A.<n> = AsymptoticRing('n^ZZ', QQ, locals={'a': 42})
        sage: A.locals()
        {'a': 42}
    """
    @staticmethod
    def _convert_locals_(locals):
        r"""
        This helper method return data converted to :class:`Locals`.

        TESTS::

            sage: A.<n> = AsymptoticRing('n^ZZ', QQ)
            sage: locals = A._convert_locals_({'a': 42}); locals
            {'a': 42}
            sage: type(locals)
            <class 'sage.rings.asymptotic.misc.Locals'>
        """
        if locals is None:
            return Locals()
        return Locals(locals)

    def locals(self, locals=None):
        r"""
        Return the actual :class:`Locals` object to be used.

        INPUT:

        - ``locals`` -- an object

          If ``locals`` is not ``None``, then a :class:`Locals` object
          is created and returned.
          If ``locals`` is ``None``, then a stored :class:`Locals` object,
          if any, is returned. Otherwise, an empty (i.e. no values except
          the default values)
          :class:`Locals` object is created and returned.

        OUTPUT: a :class:`Locals` object

        TESTS::

            sage: A.<n> = AsymptoticRing('n^ZZ', QQ, locals={'a': 42})
            sage: A.locals()
            {'a': 42}
            sage: A.locals({'a': 41})
            {'a': 41}
            sage: A.locals({'b': -3})
            {'b': -3}
        """
        if locals is None:
            try:
                locals = self._locals_
            except AttributeError:
                pass
        return self._convert_locals_(locals)



r"""
Nilpotent Lie algebras

AUTHORS:

- Eero Hakavuori (2018-08-16): initial version
"""

# ****************************************************************************
#       Copyright (C) 2018 Eero Hakavuori <eero.hakavuori@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.algebras.lie_algebras.structure_coefficients import LieAlgebraWithStructureCoefficients
from sage.categories.lie_algebras import LieAlgebras
from sage.misc.cachefunc import cached_function, cached_method
from sage.structure.global_options import GlobalOptions
from sage.structure.indexed_generators import standardize_names_index_set
from sage.rings.integer_ring import ZZ


class NilpotentLieAlgebra_dense(LieAlgebraWithStructureCoefficients):
    r"""
    A nilpotent Lie algebra `L` over a base ring.

    INPUT:

    - ``R`` -- the base ring
    - ``s_coeff`` -- dictionary of structural coefficients
    - ``names`` -- (default: ``None``) list of strings to use as names of basis
      elements; if ``None``, the names will be inferred from the structural
      coefficients
    - ``index_set`` -- (default: ``None``) list of hashable and comparable
      elements to use for indexing
    - ``step`` -- integer (optional); the nilpotency step of the
      Lie algebra if known; otherwise it will be computed when needed
    - ``category`` -- (optional) a subcategory of finite dimensional
      nilpotent Lie algebras with basis

    EXAMPLES:

    The input to a :class:`NilpotentLieAlgebra_dense` should be of the
    same form as to a
    :class:`~sage.algebras.lie_algebras.structure_coefficients.LieAlgebraWithStructureCoefficients`::

        sage: L.<X,Y,Z,W> = LieAlgebra(QQ, {('X','Y'): {'Z': 1}}, nilpotent=True)
        sage: L
        Nilpotent Lie algebra on 4 generators (X, Y, Z, W) over Rational Field
        sage: L[X, Y]
        Z
        sage: L[X, W]
        0

    If the parameter ``names`` is omitted, then the terms appearing in the
    structural coefficients are used as names::

        sage: L = LieAlgebra(QQ, {('X','Y'): {'Z': 1}}, nilpotent=True); L
        Nilpotent Lie algebra on 3 generators (X, Y, Z) over Rational Field

    TESTS::

        sage: L = LieAlgebra(QQ, {('X','Y'): {'Z': 1},
        ....:                     ('X','Z'): {'W': 1},
        ....:                     ('Y','Z'): {'T': 1}}, nilpotent=True)
        sage: TestSuite(L).run()
    """

    @staticmethod
    def __classcall_private__(cls, R, s_coeff, names=None, index_set=None,
                              category=None, **kwds):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES:

        If the variable order is specified, the order of structural
        coefficients does not matter::

            sage: from sage.algebras.lie_algebras.nilpotent_lie_algebra import NilpotentLieAlgebra_dense
            sage: L1.<x,y,z> = NilpotentLieAlgebra_dense(QQ, {('x','y'): {'z': 1}})
            sage: L2.<x,y,z> = NilpotentLieAlgebra_dense(QQ, {('y','x'): {'z': -1}})
            sage: L1 is L2
            True

        If the variables are implicitly defined by the structural coefficients,
        the ordering may be different and the Lie algebras will be considered
        different::

            sage: from sage.algebras.lie_algebras.nilpotent_lie_algebra import NilpotentLieAlgebra_dense
            sage: L1 = NilpotentLieAlgebra_dense(QQ, {('x','y'): {'z': 1}})
            sage: L2 = NilpotentLieAlgebra_dense(QQ, {('y','x'): {'z': -1}})
            sage: L1
            Nilpotent Lie algebra on 3 generators (x, y, z) over Rational Field
            sage: L2
            Nilpotent Lie algebra on 3 generators (y, x, z) over Rational Field
            sage: L1 is L2
            False

        Constructed using two different methods from :class:`LieAlgebra`
        yields the same Lie algebra::

            sage: sc = {('X','Y'): {'Z': 1}}
            sage: C = LieAlgebras(QQ).Nilpotent().FiniteDimensional().WithBasis()
            sage: L1.<X,Y,Z> = LieAlgebra(QQ, sc, category=C)
            sage: L2 = LieAlgebra(QQ, sc, nilpotent=True, names=['X','Y','Z'])
            sage: L1 is L2
            True
        """
        if not names:
            # extract names from structural coefficients
            names = []
            for (X, Y), d in s_coeff.items():
                if X not in names:
                    names.append(X)
                if Y not in names:
                    names.append(Y)
                for k in d:
                    if k not in names:
                        names.append(k)

        from sage.structure.indexed_generators import standardize_names_index_set
        names, index_set = standardize_names_index_set(names, index_set)
        s_coeff = LieAlgebraWithStructureCoefficients._standardize_s_coeff(
            s_coeff, index_set)

        cat = LieAlgebras(R).FiniteDimensional().WithBasis().Nilpotent()
        category = cat.or_subcategory(category)

        return super().__classcall__(cls, R, s_coeff, names,
                                     index_set, category=category, **kwds)

    def __init__(self, R, s_coeff, names, index_set, step=None, **kwds) -> None:
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: L.<X,Y,Z,W> = LieAlgebra(QQ, {('X','Y'): {'Z': 1}}, nilpotent=True)
            sage: TestSuite(L).run()
        """
        if step is not None:
            self._step = step

        LieAlgebraWithStructureCoefficients.__init__(self, R, s_coeff,
                                                     names, index_set,
                                                     **kwds)

    def _repr_(self) -> str:
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: L.<X,Y,Z,W> = LieAlgebra(QQ, {('X','Y'): {'Z': 1}}, nilpotent=True)
            sage: L
            Nilpotent Lie algebra on 4 generators (X, Y, Z, W) over Rational Field
        """
        return "Nilpotent %s" % (super()._repr_())


class FreeNilpotentLieAlgebra(NilpotentLieAlgebra_dense):
    r"""
    Return the free nilpotent Lie algebra of step ``s`` with ``r`` generators.

    The free nilpotent Lie algebra `L` of step `s` with `r` generators is
    the quotient of the free Lie algebra on `r` generators by the `(s+1)`-th
    term of the lower central series. That is, the only relations in the
    Lie algebra `L` are anticommutativity, the Jacobi identity, and the
    vanishing of all brackets of length more than `s`.

    INPUT:

    - ``R`` -- the base ring
    - ``r`` -- integer; the number of generators
    - ``s`` -- integer; the nilpotency step of the algebra
    - ``names`` -- (optional) string or list of strings used to name the
      basis elements; if ``names`` is a string, then names for the basis
      will be autogenerated as determined by the ``naming`` parameter
    - ``naming`` -- (optional) a string; the naming scheme to use for
      the basis; valid values are:

      * ``'index'`` -- the basis elements are ``names_w``, where ``w``
        are Lyndon words indexing the basis
      * ``'linear'`` -- the basis is indexed ``names_1``, ..., ``names_n``
        in the ordering of the Lyndon basis

    .. NOTE::

        If `10 \leq r \leq 16`, then the Lyndon words are in the
        alphabet of hexadecimals.

    EXAMPLES:

    We compute the free step 4 Lie algebra on 2 generators and
    verify the only non-trivial relation
    `[[X_1,[X_1,X_2]],X_2] = [X_1,[[X_1,X_2],X_2]]`::

        sage: L = LieAlgebra(QQ, 2, step=4)
        sage: L.basis().list()
        [X_1, X_2, X_12, X_112, X_122, X_1112, X_1122, X_1222]
        sage: X_1, X_2 = L.basis().list()[:2]
        sage: L[[X_1, [X_1, X_2]], X_2]
        X_1122
        sage: L[[X_1, [X_1, X_2]], X_2] == L[X_1, [[X_1, X_2], X_2]]
        True

    The linear naming scheme on the same Lie algebra::

        sage: K = LieAlgebra(QQ, 2, step=4, names='Y', naming='linear')
        sage: K.basis().list()
        [Y_1, Y_2, Y_3, Y_4, Y_5, Y_6, Y_7, Y_8]
        sage: K.inject_variables()
        Defining Y_1, Y_2, Y_3, Y_4, Y_5, Y_6, Y_7, Y_8
        sage: Y_2.bracket(Y_3)
        -Y_5
        sage: Y_5.bracket(Y_1)
        -Y_7
        sage: Y_3.bracket(Y_4)
        0

    A fully custom naming scheme on the Heisenberg algebra::

        sage: L = LieAlgebra(ZZ, 2, step=2, names=('X', 'Y', 'Z'))
        sage: a, b, c = L.basis()
        sage: L.basis().list()
        [X, Y, Z]
        sage: a.bracket(b)
        Z

    An equivalent way to define custom names for the basis elements and
    bind them as local variables simultaneously::

        sage: L.<X,Y,Z> = LieAlgebra(ZZ, 2, step=2)
        sage: L.basis().list()
        [X, Y, Z]
        sage: X.bracket(Y)
        Z

    A free nilpotent Lie algebra is a stratified nilpotent Lie algebra::

        sage: L = LieAlgebra(QQ, 3, step=3)
        sage: L.category()
        Category of finite dimensional stratified Lie algebras with basis over Rational Field
        sage: L in LieAlgebras(QQ).Nilpotent()
        True

    Being graded means that each basis element has a degree::

        sage: L in LieAlgebras(QQ).Graded()
        True
        sage: L.homogeneous_component_basis(1).list()
        [X_1, X_2, X_3]
        sage: L.homogeneous_component_basis(2).list()
        [X_12, X_13, X_23]
        sage: L.homogeneous_component_basis(3).list()
        [X_112, X_113, X_122, X_123, X_132, X_133, X_223, X_233]

    TESTS:

    Verify all bracket relations in the free nilpotent Lie algebra of step 5
    with 2 generators::

        sage: L = LieAlgebra(QQ, 2, step=5)
        sage: L.inject_variables()
        Defining X_1, X_2, X_12, X_112, X_122, X_1112, X_1122, X_1222,
        X_11112, X_11122, X_11212, X_11222, X_12122, X_12222
        sage: [X_1.bracket(Xk) for Xk in L.basis()]
        [0, X_12, X_112, X_1112, X_1122,
        X_11112, X_11122, X_11222, 0, 0, 0, 0, 0, 0]
        sage: [X_2.bracket(Xk) for Xk in L.basis()]
        [-X_12, 0, -X_122, -X_1122, -X_1222,
        -X_11122 + X_11212, -X_11222 - X_12122, -X_12222, 0, 0, 0, 0, 0, 0]
        sage: [X_12.bracket(Xk) for Xk in L.basis()]
        [-X_112, X_122, 0, -X_11212, X_12122, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_112.bracket(Xk) for Xk in L.basis()]
        [-X_1112, X_1122, X_11212, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_122.bracket(Xk) for Xk in L.basis()]
        [-X_1122, X_1222, -X_12122, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_1112.bracket(Xk) for Xk in L.basis()]
        [-X_11112, X_11122 - X_11212, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_1122.bracket(Xk) for Xk in L.basis()]
        [-X_11122, X_11222 + X_12122, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_1222.bracket(Xk) for Xk in L.basis()]
        [-X_11222, X_12222, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_11112.bracket(Xk) for Xk in L.basis()]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_11122.bracket(Xk) for Xk in L.basis()]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_11212.bracket(Xk) for Xk in L.basis()]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_11222.bracket(Xk) for Xk in L.basis()]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_12122.bracket(Xk) for Xk in L.basis()]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: [X_12222.bracket(Xk) for Xk in L.basis()]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    The dimensions of the smallest free nilpotent Lie algebras on
    2 and 3 generators::

        sage: l = [LieAlgebra(QQ, 2, step=k) for k in range(1, 7)]
        sage: [L.dimension() for L in l]
        [2, 3, 5, 8, 14, 23]
        sage: l = [LieAlgebra(QQ, 3, step=k) for k in range(1, 4)]
        sage: [L.dimension() for L in l]
        [3, 6, 14]

    Verify that a free nilpotent Lie algebra of step `>2` with `>10`
    generators can be created, see :issue:`27018` (see also :issue:`27069`)::

        sage: L = LieAlgebra(QQ, 11, step=3)
        sage: L.dimension() == 11 + (11^2-11)/2 + (11^3-11)/3
        True

    Verify that we can create a free nilpotent Lie algebra with a "large"
    number of generators::

        sage: list(LieAlgebra(QQ, 14, step=2).basis(1))
        [X_1, X_2, X_3, X_4, X_5, X_6, X_7, X_8, X_9, X_a, X_b, X_c, X_d, X_e]

        sage: L = LieAlgebra(QQ, 17, step=2)
        sage: len(set(repr(b) for b in L.basis())) == L.dimension()
        True
    """
    @staticmethod
    def __classcall_private__(cls, R, r, s, names=None, naming='index', category=None, **kwds):
        """
        Normalize input to ensure a unique representation.

        TESTS::

            sage: L1.<X> = LieAlgebra(ZZ, 2, step=2)
            sage: L2 = LieAlgebra(ZZ, 2, step=2, names=['X'])
            sage: L3 = LieAlgebra(ZZ, 2, step=2)
            sage: L1 is L2 and L2 is L3
            True
        """
        if names is None:
            names = 'X'

        cat = LieAlgebras(R).FiniteDimensional().WithBasis()
        category = cat.Graded().Stratified().or_subcategory(category)

        return super().__classcall__(
            cls, R, r, s, names=tuple(names), naming=naming,
            category=category, **kwds)

    def __init__(self, R, r, s, names, naming, category, **kwds) -> None:
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: L = LieAlgebra(ZZ, 2, step=2)
            sage: TestSuite(L).run()

            sage: L = LieAlgebra(QQ, 4, step=3)
            sage: TestSuite(L).run()  # long time
        """
        if r not in ZZ or r <= 0:
            raise ValueError("number of generators %s is not "
                             "a positive integer" % r)
        if s not in ZZ or s <= 0:
            raise ValueError("step %s is not a positive integer" % s)

        # extract an index set from the Lyndon words of the corresponding
        # free Lie algebra, and store the corresponding elements in a dict
        from sage.algebras.lie_algebras.lie_algebra import LieAlgebra

        free_gen_names = ['F%d' % k for k in range(r)]
        free_gen_names_inv = {val: i + 1 for i, val in enumerate(free_gen_names)}
        L = LieAlgebra(R, free_gen_names).Lyndon()

        basis_by_deg = {d: [] for d in range(1, s + 1)}
        for d in range(1, s + 1):
            for X in L.graded_basis(d):
                # convert brackets of form [X_1, [X_1, X_2]] to words (1,1,2)
                w = tuple(free_gen_names_inv[s]
                          for s in X.leading_support().to_word())
                basis_by_deg[d].append((w, X))

        index_set = [ind for d in basis_by_deg for ind, val in basis_by_deg[d]]

        if len(names) == 1 and len(index_set) > 1:
            if naming == 'linear':
                names = ['%s_%d' % (names[0], k + 1)
                         for k in range(len(index_set))]
            elif naming == 'index':
                if r < 10:
                    let = repr
                elif r <= 16:
                    hexdata = [repr(i) for i in range(10)] + ['a','b','c','d','e','f']

                    def let(i):
                        return hexdata[i]
                else:
                    rlen = len(repr(r))

                    def let(i):
                        ret = repr(i)
                        return '0'*(rlen-len(ret)) + ret
                names = ['%s_%s' % (names[0], "".join(let(s) for s in ind))
                         for ind in index_set]
            else:
                raise ValueError("unknown naming scheme %s" % naming)

        # extract structural coefficients from the free Lie algebra
        s_coeff = {}
        for dx in range(1, s + 1):
            # Brackets are only computed when deg(X) + deg(Y) <= s
            # We also require deg(Y) >= deg(X) by the ordering
            for dy in range(dx, s + 1 - dx):
                if dx == dy:
                    for i, val in enumerate(basis_by_deg[dx]):
                        X_ind, X = val
                        for Y_ind, Y in basis_by_deg[dy][i + 1:]:
                            Z = L[X, Y]
                            if not Z.is_zero():
                                s_coeff[(X_ind, Y_ind)] = {W_ind: Z[W.leading_support()]
                                                           for W_ind, W in basis_by_deg[dx + dy]}
                else:
                    for X_ind, X in basis_by_deg[dx]:
                        for Y_ind, Y in basis_by_deg[dy]:
                            Z = L[X, Y]
                            if not Z.is_zero():
                                s_coeff[(X_ind, Y_ind)] = {W_ind: Z[W.leading_support()]
                                                           for W_ind, W in basis_by_deg[dx + dy]}

        names, index_set = standardize_names_index_set(names, index_set)
        s_coeff = LieAlgebraWithStructureCoefficients._standardize_s_coeff(
            s_coeff, index_set)

        self._rank = r
        NilpotentLieAlgebra_dense.__init__(self, R, s_coeff, names,
                                           index_set, s,
                                           category=category, **kwds)

    class options(GlobalOptions):
        r"""
        Set the global options for elements of a free nilpotent Lie algebra.

        @OPTIONS@

        If no parameters are set, then the function returns a copy of the
        options dictionary.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 2, step=4)
            sage: L.an_element()
            X_1 + X_2 + X_12 + X_112 + X_122 + X_1112 + X_1122 + X_1222
            sage: L.options.display = "bracket"
            sage: L.an_element()
            X_1 + X_2 + [X_1, X_2] + [X_1, [X_1, X_2]] + [[X_1, X_2], X_2]
             + [X_1, [X_1, [X_1, X_2]]] + [X_1, [[X_1, X_2], X_2]]
             + [[[X_1, X_2], X_2], X_2]
            sage: L.options._reset()
        """
        NAME = 'FreeNilpotentLieAlgebra'
        module = 'sage.algebras.lie_algebras.nilpotent_lie_algebra'
        display = dict(default='variables',
                     description='Controls the way elements are printed',
                     values=dict(variables='print basis elements as variables',
                                 brackets='print basis elements as brackets'),
                     case_sensitive=False)

    def _repr_generator(self, w, use_latex=False) -> str:
        r"""
        Return the string representation of the basis element
        indexed by the word ``w`` in ``self``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 2, step=4)
            sage: L._repr_generator((1, 1, 2, 2))
            'X_1122'
            sage: L = LieAlgebra(QQ, 2, step=4, naming='linear')
            sage: L._repr_generator((1, 1, 2, 2))
            'X_7'
            sage: L.<X,Y,Z> = LieAlgebra(QQ, 2, step=2)
            sage: L._repr_generator((1, 2))
            'Z'

        Check that the global option for display works::

            sage: L = LieAlgebra(QQ, 4, step=3)
            sage: L.an_element()
            X_1 + X_2 + X_3 + X_4 + X_12 + X_13 + X_14 + X_23 + X_24
             + X_34 + X_112 + X_113 + X_114 + X_122 + X_123 + X_124 + X_132
             + X_133 + X_134 + X_142 + X_143 + X_144 + X_223 + X_224 + X_233
             + X_234 + X_243 + X_244 + X_334 + X_344
            sage: L.options.display = "bracket"
            sage: L.an_element()
            X_1 + X_2 + X_3 + X_4 + [X_1, X_2] + [X_1, X_3] + [X_1, X_4]
            + ... + [X_3, [X_3, X_4]] + [[X_3, X_4], X_4]
            sage: L.options._reset()
        """
        if self.options.display == "variables" or len(w) == 1:
            i = self.indices().index(w)
            ret = self.variable_names()[i]
            if use_latex:
                from sage.misc.latex import latex
                ind = ret.find("_")
                if ind != -1:
                    ret = ret[:ind] + "_{{{}}}".format(latex(ret[ind+1:]))
            return ret

        basis = self.basis()
        from sage.algebras.lie_algebras.free_lie_algebra import is_lyndon

        @cached_function
        def standard_bracket(lw):
            if len(lw) == 1:
                return basis[lw]

            for i in range(1, len(lw)):
                if is_lyndon(lw[i:]):
                    return [standard_bracket(lw[:i]), standard_bracket(lw[i:])]

        if use_latex:
            from sage.misc.latex import latex
            return latex(standard_bracket(w))
        return repr(standard_bracket(w))

    def _latex_term(self, w) -> str:
        r"""
        Return the latex representation of the basis element
        indexed by the word ``w`` in ``self``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 2, step=4)
            sage: L._latex_term((1, 1, 2, 2))
            'X_{1122}'
            sage: L = LieAlgebra(QQ, 2, step=4, naming='linear')
            sage: L._latex_term((1, 1, 2, 2))
            'X_{7}'
            sage: L.<X,Y,Z> = LieAlgebra(QQ, 2, step=2)
            sage: L._latex_term((1, 2))
            'Z'

        Check that the global option for display works::

            sage: L = LieAlgebra(QQ, 2, step=3)
            sage: latex(L.an_element())
            X_{1} + X_{2} + X_{12} + X_{112} + X_{122}
            sage: L.options.display = "bracket"
            sage: latex(L.an_element())
            X_{1} + X_{2} + \left[X_{1}, X_{2}\right]
             + \left[X_{1}, \left[X_{1}, X_{2}\right]\right]
             + \left[\left[X_{1}, X_{2}\right], X_{2}\right]
            sage: L.options._reset()
        """
        return self._repr_generator(w, use_latex=True)

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 2, step=5)
            sage: L
            Free Nilpotent Lie algebra of rank 2 and step 5 over Rational Field
        """
        return "Free Nilpotent Lie algebra of rank {} and step {} over {}".format(
            self._rank, self._step, self.base_ring())

    def degree_on_basis(self, w) -> int:
        r"""
        Return the degree of the basis element index by ``w``.

        EXAMPLES::

            sage: L = LieAlgebra(QQ, 2, step=3)
            sage: for b in L.basis():
            ....:     b, L.degree_on_basis(b.leading_support())
            (X_1, 1)
            (X_2, 1)
            (X_12, 2)
            (X_112, 3)
            (X_122, 3)
        """
        return len(w)



r"""
Catalog of matroids

A list of parametrized and individual matroids. The individual matroids are
grouped into collections.

All listed matroids are available via tab completion. Simply type ``matroids.``
+ :kbd:`Tab` to see the various constructions available. For individual
matroids type ``matroids.catalog.`` + :kbd:`Tab`.

.. NOTE::

    To create a custom matroid using a variety of inputs, see the function
    :func:`Matroid() <sage.matroids.constructor.Matroid>`.

- Parametrized matroids (``matroids.`` + :kbd:`Tab`)
    - :func:`matroids.AG <sage.matroids.database_matroids.AG>`
    - :func:`matroids.CompleteGraphic <sage.matroids.database_matroids.CompleteGraphic>`
    - :func:`matroids.PG <sage.matroids.database_matroids.PG>`
    - :func:`matroids.Psi <sage.matroids.database_matroids.Psi>`
    - :func:`matroids.Spike <sage.matroids.database_matroids.Spike>`
    - :func:`matroids.Theta <sage.matroids.database_matroids.Theta>`
    - :func:`matroids.Uniform <sage.matroids.database_matroids.Uniform>`
    - :func:`matroids.Wheel <sage.matroids.database_matroids.Wheel>`
    - :func:`matroids.Whirl <sage.matroids.database_matroids.Whirl>`
    - :func:`matroids.Z <sage.matroids.database_matroids.Z>`

- Oxley's matroid collection (``matroids.catalog.`` + :kbd:`Tab`)
    - :func:`matroids.catalog.U24 <sage.matroids.database_matroids.U24>`
    - :func:`matroids.catalog.U25 <sage.matroids.database_matroids.U25>`
    - :func:`matroids.catalog.U35 <sage.matroids.database_matroids.U35>`
    - :func:`matroids.catalog.K4 <sage.matroids.database_matroids.K4>`
    - :func:`matroids.catalog.Whirl3 <sage.matroids.database_matroids.Whirl3>`
    - :func:`matroids.catalog.Q6 <sage.matroids.database_matroids.Q6>`
    - :func:`matroids.catalog.P6 <sage.matroids.database_matroids.P6>`
    - :func:`matroids.catalog.U36 <sage.matroids.database_matroids.U36>`
    - :func:`matroids.catalog.R6 <sage.matroids.database_matroids.R6>`
    - :func:`matroids.catalog.Fano <sage.matroids.database_matroids.Fano>`
    - :func:`matroids.catalog.FanoDual <sage.matroids.database_matroids.FanoDual>`
    - :func:`matroids.catalog.NonFano <sage.matroids.database_matroids.NonFano>`
    - :func:`matroids.catalog.NonFanoDual <sage.matroids.database_matroids.NonFanoDual>`
    - :func:`matroids.catalog.O7 <sage.matroids.database_matroids.O7>`
    - :func:`matroids.catalog.P7 <sage.matroids.database_matroids.P7>`
    - :func:`matroids.catalog.AG32 <sage.matroids.database_matroids.AG32>`
    - :func:`matroids.catalog.AG32prime <sage.matroids.database_matroids.AG32prime>`
    - :func:`matroids.catalog.R8 <sage.matroids.database_matroids.R8>`
    - :func:`matroids.catalog.F8 <sage.matroids.database_matroids.F8>`
    - :func:`matroids.catalog.Q8 <sage.matroids.database_matroids.Q8>`
    - :func:`matroids.catalog.L8 <sage.matroids.database_matroids.L8>`
    - :func:`matroids.catalog.S8 <sage.matroids.database_matroids.S8>`
    - :func:`matroids.catalog.Vamos <sage.matroids.database_matroids.Vamos>`
    - :func:`matroids.catalog.T8 <sage.matroids.database_matroids.T8>`
    - :func:`matroids.catalog.J <sage.matroids.database_matroids.J>`
    - :func:`matroids.catalog.P8 <sage.matroids.database_matroids.P8>`
    - :func:`matroids.catalog.P8pp <sage.matroids.database_matroids.P8pp>`
    - :func:`matroids.catalog.Wheel4 <sage.matroids.database_matroids.Wheel4>`
    - :func:`matroids.catalog.Whirl4 <sage.matroids.database_matroids.Whirl4>`
    - :func:`matroids.catalog.K33dual <sage.matroids.database_matroids.K33dual>`
    - :func:`matroids.catalog.K33 <sage.matroids.database_matroids.K33>`
    - :func:`matroids.catalog.AG23 <sage.matroids.database_matroids.AG23>`
    - :func:`matroids.catalog.TernaryDowling3 <sage.matroids.database_matroids.TernaryDowling3>`
    - :func:`matroids.catalog.R9 <sage.matroids.database_matroids.R9>`
    - :func:`matroids.catalog.Pappus <sage.matroids.database_matroids.Pappus>`
    - :func:`matroids.catalog.NonPappus <sage.matroids.database_matroids.NonPappus>`
    - :func:`matroids.catalog.K5 <sage.matroids.database_matroids.K5>`
    - :func:`matroids.catalog.K5dual <sage.matroids.database_matroids.K5dual>`
    - :func:`matroids.catalog.R10 <sage.matroids.database_matroids.R10>`
    - :func:`matroids.catalog.NonDesargues <sage.matroids.database_matroids.NonDesargues>`
    - :func:`matroids.catalog.R12 <sage.matroids.database_matroids.R12>`
    - :func:`matroids.catalog.ExtendedTernaryGolayCode <sage.matroids.database_matroids.ExtendedTernaryGolayCode>`
    - :func:`matroids.catalog.T12 <sage.matroids.database_matroids.T12>`
    - :func:`matroids.catalog.PG23 <sage.matroids.database_matroids.PG23>`

- Brettell's matroid collection (``matroids.catalog.`` + :kbd:`Tab`)
    - :func:`matroids.catalog.RelaxedNonFano <sage.matroids.database_matroids.RelaxedNonFano>`
    - :func:`matroids.catalog.AG23minusDY <sage.matroids.database_matroids.AG23minusDY>`
    - :func:`matroids.catalog.TQ8 <sage.matroids.database_matroids.TQ8>`
    - :func:`matroids.catalog.P8p <sage.matroids.database_matroids.P8p>`
    - :func:`matroids.catalog.KP8 <sage.matroids.database_matroids.KP8>`
    - :func:`matroids.catalog.Sp8 <sage.matroids.database_matroids.Sp8>`
    - :func:`matroids.catalog.Sp8pp <sage.matroids.database_matroids.Sp8pp>`
    - :func:`matroids.catalog.LP8 <sage.matroids.database_matroids.LP8>`
    - :func:`matroids.catalog.WQ8 <sage.matroids.database_matroids.WQ8>`
    - :func:`matroids.catalog.BB9 <sage.matroids.database_matroids.BB9>`
    - :func:`matroids.catalog.TQ9 <sage.matroids.database_matroids.TQ9>`
    - :func:`matroids.catalog.TQ9p <sage.matroids.database_matroids.TQ9p>`
    - :func:`matroids.catalog.M8591 <sage.matroids.database_matroids.M8591>`
    - :func:`matroids.catalog.PP9 <sage.matroids.database_matroids.PP9>`
    - :func:`matroids.catalog.BB9gDY <sage.matroids.database_matroids.BB9gDY>`
    - :func:`matroids.catalog.A9 <sage.matroids.database_matroids.A9>`
    - :func:`matroids.catalog.FN9 <sage.matroids.database_matroids.FN9>`
    - :func:`matroids.catalog.FX9 <sage.matroids.database_matroids.FX9>`
    - :func:`matroids.catalog.KR9 <sage.matroids.database_matroids.KR9>`
    - :func:`matroids.catalog.KQ9 <sage.matroids.database_matroids.KQ9>`
    - :func:`matroids.catalog.UG10 <sage.matroids.database_matroids.UG10>`
    - :func:`matroids.catalog.FF10 <sage.matroids.database_matroids.FF10>`
    - :func:`matroids.catalog.GP10 <sage.matroids.database_matroids.GP10>`
    - :func:`matroids.catalog.FZ10 <sage.matroids.database_matroids.FZ10>`
    - :func:`matroids.catalog.UQ10 <sage.matroids.database_matroids.UQ10>`
    - :func:`matroids.catalog.FP10 <sage.matroids.database_matroids.FP10>`
    - :func:`matroids.catalog.TQ10 <sage.matroids.database_matroids.TQ10>`
    - :func:`matroids.catalog.FY10 <sage.matroids.database_matroids.FY10>`
    - :func:`matroids.catalog.PP10 <sage.matroids.database_matroids.PP10>`
    - :func:`matroids.catalog.FU10 <sage.matroids.database_matroids.FU10>`
    - :func:`matroids.catalog.D10 <sage.matroids.database_matroids.D10>`
    - :func:`matroids.catalog.UK10 <sage.matroids.database_matroids.UK10>`
    - :func:`matroids.catalog.PK10 <sage.matroids.database_matroids.PK10>`
    - :func:`matroids.catalog.GK10 <sage.matroids.database_matroids.GK10>`
    - :func:`matroids.catalog.FT10 <sage.matroids.database_matroids.FT10>`
    - :func:`matroids.catalog.TK10 <sage.matroids.database_matroids.TK10>`
    - :func:`matroids.catalog.KT10 <sage.matroids.database_matroids.KT10>`
    - :func:`matroids.catalog.TU10 <sage.matroids.database_matroids.TU10>`
    - :func:`matroids.catalog.UT10 <sage.matroids.database_matroids.UT10>`
    - :func:`matroids.catalog.FK10 <sage.matroids.database_matroids.FK10>`
    - :func:`matroids.catalog.KF10 <sage.matroids.database_matroids.KF10>`
    - :func:`matroids.catalog.FA11 <sage.matroids.database_matroids.FA11>`
    - :func:`matroids.catalog.FR12 <sage.matroids.database_matroids.FR12>`
    - :func:`matroids.catalog.GP12 <sage.matroids.database_matroids.GP12>`
    - :func:`matroids.catalog.FQ12 <sage.matroids.database_matroids.FQ12>`
    - :func:`matroids.catalog.FF12 <sage.matroids.database_matroids.FF12>`
    - :func:`matroids.catalog.FZ12 <sage.matroids.database_matroids.FZ12>`
    - :func:`matroids.catalog.UQ12 <sage.matroids.database_matroids.UQ12>`
    - :func:`matroids.catalog.FP12 <sage.matroids.database_matroids.FP12>`
    - :func:`matroids.catalog.FS12 <sage.matroids.database_matroids.FS12>`
    - :func:`matroids.catalog.UK12 <sage.matroids.database_matroids.UK12>`
    - :func:`matroids.catalog.UA12 <sage.matroids.database_matroids.UA12>`
    - :func:`matroids.catalog.AK12 <sage.matroids.database_matroids.AK12>`
    - :func:`matroids.catalog.FK12 <sage.matroids.database_matroids.FK12>`
    - :func:`matroids.catalog.KB12 <sage.matroids.database_matroids.KB12>`
    - :func:`matroids.catalog.AF12 <sage.matroids.database_matroids.AF12>`
    - :func:`matroids.catalog.NestOfTwistedCubes <sage.matroids.database_matroids.NestOfTwistedCubes>`
    - :func:`matroids.catalog.XY13 <sage.matroids.database_matroids.XY13>`
    - :func:`matroids.catalog.N3 <sage.matroids.database_matroids.N3>`
    - :func:`matroids.catalog.N3pp <sage.matroids.database_matroids.N3pp>`
    - :func:`matroids.catalog.UP14 <sage.matroids.database_matroids.UP14>`
    - :func:`matroids.catalog.VP14 <sage.matroids.database_matroids.VP14>`
    - :func:`matroids.catalog.FV14 <sage.matroids.database_matroids.FV14>`
    - :func:`matroids.catalog.OW14 <sage.matroids.database_matroids.OW14>`
    - :func:`matroids.catalog.FM14 <sage.matroids.database_matroids.FM14>`
    - :func:`matroids.catalog.FA15 <sage.matroids.database_matroids.FA15>`
    - :func:`matroids.catalog.N4 <sage.matroids.database_matroids.N4>`

- Collection of various matroids (``matroids.catalog.`` + :kbd:`Tab`)
    - :func:`matroids.catalog.NonVamos <sage.matroids.database_matroids.NonVamos>`
    - :func:`matroids.catalog.TicTacToe <sage.matroids.database_matroids.TicTacToe>`
    - :func:`matroids.catalog.Q10 <sage.matroids.database_matroids.Q10>`
    - :func:`matroids.catalog.N1 <sage.matroids.database_matroids.N1>`
    - :func:`matroids.catalog.N2 <sage.matroids.database_matroids.N2>`
    - :func:`matroids.catalog.BetsyRoss <sage.matroids.database_matroids.BetsyRoss>`
    - :func:`matroids.catalog.Block_9_4 <sage.matroids.database_matroids.Block_9_4>`
    - :func:`matroids.catalog.Block_10_5 <sage.matroids.database_matroids.Block_10_5>`
    - :func:`matroids.catalog.ExtendedBinaryGolayCode <sage.matroids.database_matroids.ExtendedBinaryGolayCode>`
    - :func:`matroids.catalog.AG23minus <sage.matroids.database_matroids.AG23minus>`
    - :func:`matroids.catalog.NotP8 <sage.matroids.database_matroids.NotP8>`
    - :func:`matroids.catalog.D16 <sage.matroids.database_matroids.D16>`
    - :func:`matroids.catalog.Terrahawk <sage.matroids.database_matroids.Terrahawk>`
    - :func:`matroids.catalog.R9A <sage.matroids.database_matroids.R9A>`
    - :func:`matroids.catalog.R9B <sage.matroids.database_matroids.R9B>`
    - :func:`matroids.catalog.P9 <sage.matroids.database_matroids.P9>`
"""

# Do not add code to this file, only imports.
# Workaround for help in the notebook (needs parentheses in this file)
from sage.misc.lazy_import import lazy_import

# user-accessible:
lazy_import(
    "sage.matroids.database_collections",
    ("AllMatroids", "BrettellMatroids", "OxleyMatroids", "VariousMatroids"),
)
lazy_import(
    "sage.matroids.database_matroids",
    (
        "AG",
        "CompleteGraphic",
        "PG",
        "Psi",
        "Spike",
        "Theta",
        "Uniform",
        "Wheel",
        "Whirl",
        "Z",
    ),
)
lazy_import("sage.matroids", "catalog")

del lazy_import  # delete so it doesn't appear under tab completion



r"""
Jacobian of a general hyperelliptic curve

AUTHORS:

- David Kohel (2006): initial version
- Sabrina Kunzweiler, Gareth Ma, Giacomo Pope (2024): adapt to smooth model
"""

# ****************************************************************************
#       Copyright (C) 2025 Sabrina Kunzweiler <sabrina.kunzweiler@math.u-bordeaux.fr>
#                     2025 Gareth Ma <grhkm21@gmail.com>
#                     2025 Giacomo Pope <giacomopope@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.rings.integer import Integer
from sage.rings.rational_field import QQ
from sage.schemes.hyperelliptic_curves import (
    jacobian_homset_inert,
    jacobian_homset_ramified,
    jacobian_homset_split,
    jacobian_morphism,
)
from sage.schemes.jacobians.abstract_jacobian import Jacobian_generic


class HyperellipticJacobian_generic(Jacobian_generic):
    r"""
    This is the base class for Jacobians of hyperelliptic curves.

    We represent elements of the Jacobian by tuples of the form
    `(u, v : n)`, where

    - `(u,v)` is the Mumford representative of a divisor `P_1 + ... + P_r`,

    - `n` is a non-negative integer

    This tuple represents the equivalence class

    .. MATH::

        [P_1 + ... + P_r + n \cdot \infty_+ + m\cdot \infty_- - D_\infty],

    where  `m = g - \deg(u) - n`, and `\infty_+`, `\infty_-` are the
    points at infinity of the hyperelliptic curve,

    .. MATH::
        D_\infty =
        \lceil g/2 \rceil \infty_+ + \lfloor g/2 \rfloor \infty_-.

    Here, `\infty_- = \infty_+`, if the hyperelliptic curve is ramified.

    Such a representation exists and is unique, unless the genus `g` is odd
    and the curve is inert.

    If the hyperelliptic curve is ramified or inert, then `n` can be deduced
    from `\deg(u)` and `g`. In these cases, `n` is omitted in the description.
    """

    def dimension(self) -> Integer:
        r"""
        Return the dimension of this Jacobian.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^2, x^4+1); H
            Hyperelliptic Curve over Rational Field defined by y^2 + (x^4 + 1)*y = x^2
            sage: J = Jacobian(H)
            sage: J.dimension()
            3
        """
        return self.curve().genus()

    def is_parent_of(self, element):
        r"""
        Return whether ``self`` is the parent of ``element``.

        TESTS::

            sage: R.<x> = GF(19)[]
            sage: H = HyperellipticCurve(x^5 + x, x^2 + 1)
            sage: J = H.jacobian()
            sage: J.is_parent_of(J.zero())
            True
        """
        from sage.structure.element import parent as get_parent
        p = get_parent(element)
        return p == self or p == self.point_homset()

    def point(self, *mumford, check=True, **kwargs):
        r"""
        Return a point on the Jacobian, given:

        1. No arguments or the integer `0`; return `0 \in J`;

        2. A point `P` on `J = Jac(C)`, return `P`;

        3. A point `P` on the curve `H` such that `J = Jac(H)`;
           return `[P - P_0]`, where `P_0` is the distinguished point of `H`.
           By default, `P_0 = \infty`;

        4. Two points `P, Q` on the curve `H` such that `J = Jac(H)`;
           return `[P - Q]`;

        5. Polynomials `(u, v)` such that `v^2 + hv - f \equiv 0 \pmod u`;
           return `[(u(x), y - v(x))]`.

        EXAMPLES::

            sage: x = polygen(GF(101))
            sage: C = HyperellipticCurve(x^5 + x)
            sage: J = C.jacobian()
            sage: J.point(0)
            (1, 0)
            sage: J.point(C(-5, 1))
            (x + 5, 1)
            sage: J.point(C(-5, 1), C(0, 0))
            (x^2 + 5*x, 20*x)
            sage: J.point(C(-5, 1), C(0, 0)) - J.point(C(0, 0))
            (x + 5, 1)
            sage: J.point(x, 0)
            (x, 0)

        .. SEEALSO::

            :class:`sage.schemes.hyperelliptic_curves.jacobian_homset_generic.HyperellipticJacobianHomset`.
        """
        try:
            return self.point_homset()(*mumford, check=check)
        except AttributeError:
            raise ValueError("Arguments must determine a valid Mumford divisor.")

    def _point_homset(self, *args, **kwds):
        r"""
        Create the Hom-Set of the Jacobian according to the type of `self`.
        """
        # TODO: make a constructor for this??
        H = self.curve()
        if H.is_ramified():
            return jacobian_homset_ramified.HyperellipticJacobianHomsetRamified(
                *args, **kwds
            )
        if H.is_split():
            return jacobian_homset_split.HyperellipticJacobianHomsetSplit(*args, **kwds)
        return jacobian_homset_inert.HyperellipticJacobianHomsetInert(*args, **kwds)

    def _point(self, *args, **kwds):
        H = self.curve()
        if H.is_ramified():
            return jacobian_morphism.MumfordDivisorClassFieldRamified(*args, **kwds)
        if H.is_split():
            return jacobian_morphism.MumfordDivisorClassFieldSplit(*args, **kwds)
        return jacobian_morphism.MumfordDivisorClassFieldInert(*args, **kwds)

    @cached_method
    def order(self):
        r"""
        Compute the order of the Jacobian.

        EXAMPLES::

            sage: R.<x> = GF(3663031)[]
            sage: HyperellipticCurve(x^5 + 1758294*x^4 + 1908793*x^3 + 3033920*x^2 + 3445698*x + 3020661).jacobian().cardinality()
            13403849798842

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.jacobian_homset_generic.order`.
        """
        return self.point_homset().order()

    cardinality = order

    def count_points(self, *args, **kwds):
        r"""
        Compute the number of points on this Jacobian, possibly also over extension fields.

        EXAMPLES::

            sage: R.<x> = GF(3663031)[]
            sage: J = HyperellipticCurve(x^5 + 1758294*x^4 + 1908793*x^3 + 3033920*x^2 + 3445698*x + 3020661).jacobian()
            sage: J.count_points() == J.order()
            True
            sage: J.count_points(3)
            [13403849798842,
             180037321758127724648694244,
             2415703144684957707144219559418205805522]

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.jacobian_homset_generic.count_points`.
        """
        return self.point_homset().count_points(*args, **kwds)

    def lift_u(self, *args, **kwds):
        r"""
        Return one or all points with given `u`-coordinate.

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.jacobian_homset_generic.lift_u`.
        """
        return self.point_homset().lift_u(*args, **kwds)

    def random_element(self, *args, **kwds):
        r"""
        Return a random element of the Jacobian.

        EXAMPLES::

            sage: x = polygen(GF(11))
            sage: J = HyperellipticCurve(x^5 + x + 1).jacobian()
            sage: elt = J.random_element(); elt  # random
            (x^2 + 3*x + 10, 1)
            sage: elt in J
            True

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.jacobian_homset_generic.random_element`.
        """
        return self.point_homset().random_element(*args, **kwds)

    def some_elements(self) -> list[jacobian_morphism.MumfordDivisorClassField]:
        r"""
        Return some rational points on the Jacobian.

        EXAMPLES::

            sage: x = polygen(GF(11))
            sage: J = HyperellipticCurve(x^5 + x + 1).jacobian()
            sage: elts = J.some_elements(); elts  # random
            [(1, 0), (x^2 + 7*x + 3, 10*x + 7), (x^2 + 10*x, 5*x + 1), (x^2 + 8*x + 10, 1)]
            sage: all(elt in J for elt in elts)
            True
        """
        return [self.zero()] + [self.random_element(), self.random_element(), self.random_element()]

    def points(self, *args, **kwds):
        r"""
        Return all rational points on the Jacobian.

        EXAMPLES::

            sage: x = polygen(GF(11))
            sage: C = HyperellipticCurve(x^5 + x + 1)
            sage: J = C.jacobian()
            sage: pts = J.points()
            sage: sorted(pts)
            [(1, 0), (x, 1), (x, 10), (x + 2, 0), (x + 8, 4),
             ..., (x^2 + 10*x + 6, 3*x), (x^2 + 10*x + 6, 8*x),
             (x^2 + 10*x + 7, 5*x + 6), (x^2 + 10*x + 7, 6*x + 5)]
            sage: len(pts) == J.order()
            True

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.jacobian_homset_generic.points`.
        """

        return self.point_homset().points(*args, **kwds)

    list = points

    def __iter__(self):
        r"""
        Return an iterator over the elements of the Jacobian.
        """
        yield from self.point_homset().points()

    rational_points = points

    def abelian_group(self, *args, **kwds):
        r"""
        Return the group of rational points on this Jacobian as an
        :class:`~sage.groups.additive_abelian.additive_abelian_wrapper.AdditiveAbelianGroupWrapper`
        object.

        EXAMPLES::

            sage: x = polygen(GF(11^4))
            sage: C = HyperellipticCurve(x^5 + x + 1)
            sage: J = C.jacobian()
            sage: J.abelian_group()
            Additive abelian group isomorphic to Z/1677896 + Z/8 + Z/4 + Z/4
              embedded in Abelian group of points over Finite Field in z4 of size 11^4
                on Jacobian of Hyperelliptic Curve over Finite Field in z4 of size 11^4 defined by y^2 = x^5 + x + 1

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.jacobian_homset_generic.abelian_group`.
        """
        return self.point_homset().abelian_group(*args, **kwds)

    ####################################################################
    # Some properties of geometric Endomorphism ring and algebra
    ####################################################################

    @lazy_attribute
    def _have_established_geometrically_trivial(self):
        r"""
        Initialize the flag which determines whether or not we have
        already established if the geometric endomorphism ring is
        trivial.

        This is related to the warning at the top of the
        ``jacobian_endomorphism_utils.py`` module.

        INPUT:

        - ``self`` -- the Jacobian

        OUTPUT: the boolean ``False``; this will be updated by other methods

        EXAMPLES:

        This is LMFDB curve 262144.d.524288.2::

            sage: R.<x> = QQ[]
            sage: f = x^5 + x^4 + 4*x^3 + 8*x^2 + 5*x + 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J._have_established_geometrically_trivial
            False
        """
        return False

    @lazy_attribute
    def _have_established_geometrically_field(self):
        r"""
        Initialize the flag which determines whether or not we have
        already established if the geometric endomorphism ring is
        trivial.

        This is related to the warning at the top of the
        ``jacobian_endomorphism_utils.py`` module.

        INPUT:

        - ``self`` -- the Jacobian

        OUTPUT: the boolean ``False``; this will be updated by other methods

        EXAMPLES:

        This is LMFDB curve 262144.d.524288.2::

            sage: R.<x> = QQ[]
            sage: f = x^5 + x^4 + 4*x^3 + 8*x^2 + 5*x + 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J._have_established_geometrically_field
            False
        """
        return False

    def geometric_endomorphism_algebra_is_field(self, B=200, proof=False):
        r"""
        Return whether the geometric endomorphism algebra is a field.

        This implies that the Jacobian of the curve is geometrically
        simple. It is based on Algorithm 4.10 from [Lom2019]_

        INPUT:

        - ``B`` -- (default: 200) the bound which appears in the statement of
          the algorithm from [Lom2019]_

        - ``proof`` -- boolean (default: ``False``); whether or not to insist
          on a provably correct answer. This is related to the warning in the
          docstring of this module: if this function returns ``False``, then
          strictly speaking this has not been proven to be ``False`` until one
          has exhibited a non-trivial endomorphism, which these methods are not
          designed to carry out. If one is convinced that this method should
          return ``True``, but it is returning ``False``, then this can be
          exhibited by increasing `B`.

        OUTPUT:

        Boolean indicating whether or not the geometric endomorphism
        algebra is a field.

        EXAMPLES:

        This is LMFDB curve 262144.d.524288.2 which has QM. Although its
        Jacobian is geometrically simple, the geometric endomorphism algebra
        is not a field::

            sage: R.<x> = QQ[]
            sage: f = x^5 + x^4 + 4*x^3 + 8*x^2 + 5*x + 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_algebra_is_field()
            False

        This is LMFDB curve 50000.a.200000.1::

            sage: f = 8*x^5 + 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_algebra_is_field()
            True
        """
        from sage.interfaces.genus2reduction import genus2reduction

        from .jacobian_endomorphism_utils import get_is_geom_field

        if self._have_established_geometrically_field:
            return True
        C = self.curve()
        if C.genus() != 2:
            raise NotImplementedError(
                "Current implementation requires the curve to be of genus 2"
            )
        if C.base_ring() != QQ:
            raise NotImplementedError(
                "Current implementation requires the curve to be defined over the rationals"
            )
        f, h = C.hyperelliptic_polynomials()
        if h != 0:
            raise NotImplementedError(
                "Current implementation requires the curve to be in the form y^2 = f(x)"
            )
        red_data = genus2reduction(0, f)
        cond_C = red_data.conductor  # WARNING: this is only the prime_to_2 conductor.
        bad_primes = cond_C.prime_divisors()
        self._bad_primes = bad_primes

        is_abs_simp, is_def_geom_trivial = get_is_geom_field(f, C, bad_primes, B)

        if is_def_geom_trivial:
            self._have_established_geometrically_trivial = True
        if is_abs_simp:
            self._have_established_geometrically_field = True
            return True
        if proof:
            raise NotImplementedError(
                "Rigorous computation of lower bounds of endomorphism algebras has not yet been implemented."
            )
        return False

    def geometric_endomorphism_ring_is_ZZ(self, B=200, proof=False):
        r"""
        Return whether the geometric endomorphism ring of ``self`` is the
        integer ring `\ZZ`.

        INPUT:

        - ``B`` -- (default: 200) the bound which appears in the statement of
          the algorithm from [Lom2019]_

        - ``proof`` -- boolean (default: ``False``); whether or not to insist
          on a provably correct answer. This is related to the warning in the
          module docstring of `jacobian_endomorphisms.py`: if this function
          returns ``False``, then strictly speaking this has not been proven to
          be ``False`` until one has exhibited a non-trivial endomorphism,
          which the methods in that module are not designed to carry out. If
          one is convinced that this method should return ``True``, but it is
          returning ``False``, then this can be exhibited by increasing `B`.

        OUTPUT:

        Boolean indicating whether or not the geometric endomorphism
        ring is isomorphic to the integer ring.

        EXAMPLES:

        This is LMFDB curve 603.a.603.2::

            sage: R.<x> = QQ[]
            sage: f = 4*x^5 + x^4 - 4*x^3 + 2*x^2 + 4*x + 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_ring_is_ZZ()
            True

        This is LMFDB curve 1152.a.147456.1 whose geometric endomorphism ring
        is isomorphic to the group of 2x2 matrices over `\QQ`::

            sage: f = x^6 - 2*x^4 + 2*x^2 - 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_ring_is_ZZ()
            False

        This is LMFDB curve 20736.k.373248.1 whose geometric endomorphism ring
        is isomorphic to the group of 2x2 matrices over a CM field::

            sage: f = x^6 + 8
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_ring_is_ZZ()
            False

        This is LMFDB curve 708.a.181248.1::

            sage: R.<x> = QQ[]
            sage: f = -3*x^6 - 16*x^5 + 36*x^4 + 194*x^3 - 164*x^2 - 392*x - 143
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_ring_is_ZZ()
            True

        This is LMFDB curve 10609.a.10609.1 whose geometric endomorphism ring
        is an order in a real quadratic field::

            sage: f = x^6 + 2*x^4 + 2*x^3 + 5*x^2 + 6*x + 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_ring_is_ZZ()
            False

        This is LMFDB curve 160000.c.800000.1 whose geometric endomorphism ring
        is an order in a CM field::

            sage: f = x^5 - 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_ring_is_ZZ()
            False

        This is LMFDB curve 262144.d.524288.2 whose geometric endomorphism ring
        is an order in a quaternion algebra::

            sage: f = x^5 + x^4 + 4*x^3 + 8*x^2 + 5*x + 1
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_ring_is_ZZ()
            False

        This is LMFDB curve 578.a.2312.1 whose geometric endomorphism ring
        is `\QQ \times \QQ`::

            sage: f = 4*x^5 - 7*x^4 + 10*x^3 - 7*x^2 + 4*x
            sage: C = HyperellipticCurve(f)
            sage: J = C.jacobian()
            sage: J.geometric_endomorphism_ring_is_ZZ()
            False
        """
        from .jacobian_endomorphism_utils import is_geom_trivial_when_field

        if self._have_established_geometrically_trivial:
            return True
        is_abs_simple = self.geometric_endomorphism_algebra_is_field(B=B, proof=proof)
        if self._have_established_geometrically_trivial:
            return True
        if is_abs_simple and is_geom_trivial_when_field(self.curve(), self._bad_primes):
            return True
        if proof:
            raise NotImplementedError(
                "Rigorous computation of lower bounds of endomorphism rings has not yet been implemented."
            )
        return False



# sage.doctest: needs sage.symbolic
r"""
A Sample Session using SymPy

In this first part, we do all of the examples in the SymPy tutorial
(https://github.com/sympy/sympy/wiki/Tutorial), but using Sage
instead of SymPy.

::

    sage: a = Rational((1,2))
    sage: a
    1/2
    sage: a*2
    1
    sage: Rational(2)^50 / Rational(10)^50
    1/88817841970012523233890533447265625
    sage: 1.0/2
    0.500000000000000
    sage: 1/2
    1/2
    sage: pi^2
    pi^2
    sage: float(pi)
    3.141592653589793
    sage: RealField(200)(pi)
    3.1415926535897932384626433832795028841971693993751058209749
    sage: float(pi + exp(1))
    5.85987448204883...
    sage: oo != 2
    True

::

    sage: var('x y')
    (x, y)
    sage: x + y + x - y
    2*x
    sage: (x+y)^2
    (x + y)^2
    sage: ((x+y)^2).expand()
    x^2 + 2*x*y + y^2
    sage: ((x+y)^2).subs(x=1)
    (y + 1)^2
    sage: ((x+y)^2).subs(x=y)
    4*y^2

::

    sage: limit(sin(x)/x, x=0)
    1
    sage: limit(x, x=oo)
    +Infinity
    sage: limit((5^x + 3^x)^(1/x), x=oo)
    5

::

    sage: diff(sin(x), x)
    cos(x)
    sage: diff(sin(2*x), x)
    2*cos(2*x)
    sage: diff(tan(x), x)
    tan(x)^2 + 1
    sage: limit((tan(x+y) - tan(x))/y, y=0)
    tan(x)^2 + 1
    sage: diff(sin(2*x), x, 1)
    2*cos(2*x)
    sage: diff(sin(2*x), x, 2)
    -4*sin(2*x)
    sage: diff(sin(2*x), x, 3)
    -8*cos(2*x)

::

    sage: cos(x).taylor(x,0,10)
    -1/3628800*x^10 + 1/40320*x^8 - 1/720*x^6 + 1/24*x^4 - 1/2*x^2 + 1
    sage: (1/cos(x)).taylor(x,0,10)
    50521/3628800*x^10 + 277/8064*x^8 + 61/720*x^6 + 5/24*x^4 + 1/2*x^2 + 1

::

    sage: matrix([[1,0], [0,1]])
    [1 0]
    [0 1]
    sage: var('x y')
    (x, y)
    sage: A = matrix([[1,x], [y,1]])
    sage: A
    [1 x]
    [y 1]
    sage: A^2
    [x*y + 1     2*x]
    [    2*y x*y + 1]
    sage: R.<x,y> = QQ[]
    sage: A = matrix([[1,x], [y,1]])
    sage: A^10
    [x^5*y^5 + 45*x^4*y^4 + 210*x^3*y^3 + 210*x^2*y^2 + 45*x*y + 1     10*x^5*y^4 + 120*x^4*y^3 + 252*x^3*y^2 + 120*x^2*y + 10*x]
    [    10*x^4*y^5 + 120*x^3*y^4 + 252*x^2*y^3 + 120*x*y^2 + 10*y x^5*y^5 + 45*x^4*y^4 + 210*x^3*y^3 + 210*x^2*y^2 + 45*x*y + 1]
    sage: var('x y')
    (x, y)

And here are some actual tests of sympy::

    sage: from sympy import Symbol, cos, sympify, pprint                                # needs sympy
    sage: from sympy.abc import x                                                       # needs sympy

::

    sage: e = (1/cos(x)^3)._sympy_(); e                                                 # needs sympy
    cos(x)**(-3)
    sage: f = e.series(x, 0, int(10)); f                                                # needs sympy
    1 + 3*x**2/2 + 11*x**4/8 + 241*x**6/240 + 8651*x**8/13440 + O(x**10)

And the pretty-printer.  Since unicode characters are not working on
some architectures, we disable it::

    sage: from sympy.printing import pprint_use_unicode                                 # needs sympy
    sage: prev_use = pprint_use_unicode(False)                                          # needs sympy
    sage: pprint(e)                                                                     # needs sympy
       1
    -------
       3
    cos (x)

    sage: pprint(f)                                                                     # needs sympy
           2       4        6         8
        3*x    11*x    241*x    8651*x     / 10\
    1 + ---- + ----- + ------ + ------- + O\x  /
         2       8      240      13440
    sage: pprint_use_unicode(prev_use)                                                  # needs sympy
    False

And the functionality to convert from sympy format to Sage format::

    sage: e._sage_()                                                                    # needs sympy
    cos(x)^(-3)
    sage: e._sage_().taylor(x._sage_(), 0, 8)                                           # needs sympy
    8651/13440*x^8 + 241/240*x^6 + 11/8*x^4 + 3/2*x^2 + 1
    sage: f._sage_()                                                                    # needs sympy
    8651/13440*x^8 + 241/240*x^6 + 11/8*x^4 + 3/2*x^2 + Order(x^10) + 1

Mixing SymPy with Sage::

    sage: # needs sympy
    sage: import sympy
    sage: var("x")._sympy_() + var("y")._sympy_()
    x + y
    sage: o = var("omega")
    sage: s = sympy.Symbol("x")
    sage: t1 = s + o
    sage: t2 = o + s
    sage: type(t1)
    <class 'sympy.core.add.Add'>
    sage: type(t2)
    <class 'sage.symbolic.expression.Expression'>
    sage: t1, t2
    (omega + x, omega + x)
    sage: e = sympy.sin(var("y")) + sage.functions.trig.cos(sympy.Symbol("x"))
    sage: type(e)
    <class 'sympy.core.add.Add'>
    sage: e
    sin(y) + cos(x)
    sage: e=e._sage_()
    sage: type(e)
    <class 'sage.symbolic.expression.Expression'>
    sage: e
    cos(x) + sin(y)
    sage: e = sage.functions.trig.cos(var("y")**3)**4+var("x")**2
    sage: e = e._sympy_()
    sage: e
    x**2 + cos(y**3)**4

::

    sage: a = sympy.Matrix([1, 2, 3])                                                   # needs sympy
    sage: a[1]                                                                          # needs sympy
    2

::

    sage: sympify(1.5)                                                                  # needs sympy
    1.50000000000000
    sage: sympify(2)                                                                    # needs sympy
    2
    sage: sympify(-2)                                                                   # needs sympy
    -2

TESTS:

This was fixed in Sympy, see :issue:`14437`::

    sage: # needs sympy
    sage: from sympy import Function, Symbol, rsolve
    sage: u = Function('u')
    sage: n = Symbol('n', integer=True)
    sage: f = u(n+2) - u(n+1) + u(n)/4
    sage: expand(2**n * rsolve(f,u(n)))
    2*C1*n + C0
"""



r"""
Elements with labels.

This module implements a simple wrapper class for pairs consisting of
an "element" and a "label".
For representation purposes (``repr``, ``str``, ``latex``), this pair
behaves like its label, while the element is "silent".
However, these pairs compare like usual pairs (i.e., both element and
label have to be equal for two such pairs to be equal).
This is used for visual representations of graphs and posets
with vertex labels.
"""

from sage.misc.latex import latex


class ElementWithLabel:
    """
    Auxiliary class for showing/viewing :class:`Poset`s with
    non-injective labelings.
    For hashing and equality testing the resulting object behaves
    like a tuple ``(element, label)``.
    For any presentation purposes it appears just as ``label`` would.

    EXAMPLES::

        sage: # needs sage.combinat sage.graphs
        sage: P = Poset({1: [2,3]})
        sage: labs = {i: P.rank(i) for i in range(1, 4)}
        sage: print(labs)
        {1: 0, 2: 1, 3: 1}
        sage: print(P.plot(element_labels=labs))                                        # needs sage.plot
        Graphics object consisting of 6 graphics primitives

        sage: # needs sage.combinat sage.graphs sage.modules
        sage: from sage.misc.element_with_label import ElementWithLabel
        sage: W = WeylGroup("A1")
        sage: P = W.bruhat_poset(facade=True)
        sage: D = W.domain()
        sage: v = D.rho() - D.fundamental_weight(1)
        sage: nP = P.relabel(lambda w: ElementWithLabel(w, w.action(v)))
        sage: list(nP)
        [(0, 0), (0, 0)]
    """
    def __init__(self, element, label):
        """
        Construct an object that wraps ``element`` but presents itself
        as ``label``.

        TESTS::

            sage: from sage.misc.element_with_label import ElementWithLabel
            sage: e = ElementWithLabel(1, 'a')
            sage: e
            'a'
            sage: e.element
            1
        """
        self.element = element
        self.label = label

    def _latex_(self):
        """
        Return the latex representation of ``self``,
        which is just the latex representation of the label.

        TESTS::

            sage: var('a_1')                                                            # needs sage.symbolic
            a_1
            sage: from sage.misc.element_with_label import ElementWithLabel
            sage: e = ElementWithLabel(1, a_1)                                          # needs sage.symbolic
            sage: latex(e)                                                              # needs sage.symbolic
            a_{1}
        """
        return latex(self.label)

    def __str__(self):
        """
        Return the string representation of ``self``, which is just
        the string representation of the label.

        TESTS::

            sage: var('a_1')                                                            # needs sage.symbolic
            a_1
            sage: from sage.misc.element_with_label import ElementWithLabel
            sage: e = ElementWithLabel(1, a_1)                                          # needs sage.symbolic
            sage: str(e)                                                                # needs sage.symbolic
            'a_1'
        """
        return str(self.label)

    def __repr__(self):
        """
        Return the representation of ``self``, which is just
        the representation of the label.

        TESTS::

            sage: var('a_1')                                                            # needs sage.symbolic
            a_1
            sage: from sage.misc.element_with_label import ElementWithLabel
            sage: e = ElementWithLabel(1, a_1)                                          # needs sage.symbolic
            sage: repr(e)                                                               # needs sage.symbolic
            'a_1'
        """
        return repr(self.label)

    def __hash__(self):
        """
        Return the hash of the labeled element ``self``,
        which is just the hash of ``self.element``.

        TESTS::

            sage: from sage.misc.element_with_label import ElementWithLabel
            sage: a = ElementWithLabel(1, 'a')
            sage: b = ElementWithLabel(1, 'b')
            sage: d = {}
            sage: d[a] = 'element 1'
            sage: d[b] = 'element 2'
            sage: print(d)
            {'a': 'element 1', 'b': 'element 2'}
            sage: a = ElementWithLabel("a", [2,3])
            sage: hash(a) == hash(a.element)
            True
        """
        return hash(self.element)

    def __eq__(self, other):
        """
        Two labeled elements are equal if and only if both of their
        constituents are equal.

        TESTS::

            sage: from sage.misc.element_with_label import ElementWithLabel
            sage: a = ElementWithLabel(1, 'a')
            sage: b = ElementWithLabel(1, 'b')
            sage: x = ElementWithLabel(1, 'a')
            sage: a == b
            False
            sage: a == x
            True
            sage: 1 == a
            False
            sage: b == 1
            False
        """
        if not (isinstance(self, ElementWithLabel) and
                isinstance(other, ElementWithLabel)):
            return False
        return self.element == other.element and self.label == other.label

    def __ne__(self, other):
        """
        Two labeled elements are not equal if and only if first or second
        constituents are not equal.

        TESTS::

            sage: from sage.misc.element_with_label import ElementWithLabel
            sage: a = ElementWithLabel(1, 'a')
            sage: b = ElementWithLabel(1, 'b')
            sage: x = ElementWithLabel(1, 'a')
            sage: a != b
            True
            sage: a != x
            False
        """
        return not (self == other)



"""
Convert PARI objects to/from GMP objects

Utility function to convert PARI ``GEN``s to/from the GMP types
``mpz_t`` and ``mpq_t``.

AUTHORS:

- Luca De Feo (2016-09-06): Separate Sage-specific components from
  generic C-interface in ``Pari`` (:issue:`20241`)
"""

#*****************************************************************************
#       Copyright (C) 2016 Luca De Feo <luca.defeo@polytechnique.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from cysignals.signals cimport sig_on

from sage.libs.gmp.all cimport *

from cypari2.paridecl cimport *
from cypari2.stack cimport new_gen

cdef Gen new_gen_from_mpz_t(mpz_srcptr value):
    """
    Create a new PARI Gen of type ``t_INT`` from a given
    GMP integer ``value``.

    EXAMPLES::

        sage: pari(42)       # indirect doctest
        42

    TESTS:

    Check that the hash of an integer does not depend on existing
    garbage on the stack (:issue:`11611`)::

        sage: foo = pari(2^(32*1024))  # Create large integer to put PARI stack in known state
        sage: a5 = pari(5)
        sage: foo = pari(0xDEADBEEF * (2^(32*1024)-1)//(2^32 - 1))  # Dirty PARI stack
        sage: b5 = pari(5)
        sage: a5.__hash__() == b5.__hash__()
        True
    """
    sig_on()
    return new_gen(_new_GEN_from_mpz_t(value))


cdef inline GEN _new_GEN_from_mpz_t(mpz_srcptr value) noexcept:
    r"""
    Create a new PARI ``t_INT`` from a ``mpz_srcptr``.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    if mpz_sgn(value) == 0:
        return gen_0

    cdef unsigned long limbs = mpz_size(value)

    cdef GEN z = cgeti(limbs + 2)
    # Set sign and "effective length"
    z[1] = evalsigne(mpz_sgn(value)) + evallgefint(limbs + 2)
    mpz_export(int_LSW(z), NULL, -1, sizeof(long), 0, 0, value)

    return z


cdef Gen new_gen_from_mpq_t(mpq_t value):
    """
    Create a new PARI Gen of type ``t_INT`` or ``t_FRAC`` from a given
    GMP rational ``value``.

    EXAMPLES::

        sage: pari(-2/3)
        -2/3
        sage: pari(QQ(42))
        42
        sage: pari(QQ(42)).type()
        't_INT'
        sage: pari(QQ(1/42)).type()
        't_FRAC'

    TESTS:

    Check that the hash of a rational does not depend on existing
    garbage on the stack (:issue:`11854`)::

        sage: foo = pari(2^(32*1024))  # Create large integer to put PARI stack in known state
        sage: a5 = pari(5/7)
        sage: foo = pari(0xDEADBEEF * (2^(32*1024)-1)//(2^32 - 1))  # Dirty PARI stack
        sage: b5 = pari(5/7)
        sage: a5.__hash__() == b5.__hash__()
        True
    """
    sig_on()
    return new_gen(_new_GEN_from_mpq_t(value))


cdef inline GEN _new_GEN_from_mpq_t(mpq_t value) noexcept:
    r"""
    Create a new PARI ``t_INT`` or ``t_FRAC`` from a ``mpq_t``.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    cdef GEN num = _new_GEN_from_mpz_t(mpq_numref(value))
    if mpz_cmpabs_ui(mpq_denref(value), 1) == 0:
        # Denominator is 1, return the numerator (an integer)
        return num
    cdef GEN denom = _new_GEN_from_mpz_t(mpq_denref(value))
    return mkfrac(num, denom)


cdef Gen new_gen_from_padic(long ordp, long relprec,
                            mpz_srcptr prime, mpz_srcptr p_pow, mpz_srcptr unit):
    """
    Create a new PARI Gen of type ``t_PADIC`` from the given input data
    as GMP integers.
    """
    cdef GEN z
    sig_on()
    z = cgetg(5, t_PADIC)
    z[1] = evalprecp(relprec) + evalvalp(ordp)
    set_gel(z, 2, _new_GEN_from_mpz_t(prime))
    set_gel(z, 3, _new_GEN_from_mpz_t(p_pow))
    set_gel(z, 4, _new_GEN_from_mpz_t(unit))
    return new_gen(z)


cdef GEN _new_GEN_from_mpq_t_matrix(mpq_t** B, long nr, long nc) noexcept:
    """
    Create a new PARI ``t_MAT`` from a given
    2-dimensional array of GMP rationals ``mpq_t``.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    cdef GEN x
    # Allocate zero matrix
    cdef GEN A = zeromatcopy(nr, nc)
    cdef long i, j
    for i in range(nr):
        for j in range(nc):
            x = _new_GEN_from_mpq_t(B[i][j])
            set_gcoeff(A, i+1, j+1, x)  # A[i+1, j+1] = x (using 1-based indexing)
    return A


cdef Gen rational_matrix(mpq_t** B, long nr, long nc):
    """
    Create a new PARI matrix of type ``t_MAT`` from a given
    array of GMP rationals ``mpq_t``.

    INPUT:

    - ``B`` -- a 2-dimensional array of ``mpq_t`` values. This array is
      accessed as ``B[i][j]``, where ``i`` is the row index and ``j``
      the column index.

    - ``nr`` -- number of rows of this matrix

    - ``nc`` -- number of columns of this matrix

    EXAMPLES::

        sage: matrix(QQ,2,[1..6]).__pari__()   # indirect doctest                       # needs sage.modules
        [1, 2, 3; 4, 5, 6]
    """
    sig_on()
    cdef GEN g = _new_GEN_from_mpq_t_matrix(B, nr, nc)
    return new_gen(g)


cdef inline void INT_to_mpz(mpz_ptr value, GEN g) noexcept:
    """
    Convert a PARI ``t_INT`` to a GMP integer, stored in ``value``.
    """
    if typ(g) != t_INT:
        pari_err(e_TYPE, <char*>"conversion to mpz", g)

    cdef long size = lgefint(g) - 2
    mpz_import(value, size, -1, sizeof(long), 0, 0, int_LSW(g))

    if signe(g) < 0:
        mpz_neg(value, value)


cdef void INTFRAC_to_mpq(mpq_ptr value, GEN g) noexcept:
    """
    Convert a PARI ``t_INT`` or ``t_FRAC`` to a GMP rational, stored in
    ``value``.
    """
    if typ(g) == t_FRAC:
        INT_to_mpz(mpq_numref(value), gel(g, 1))
        INT_to_mpz(mpq_denref(value), gel(g, 2))
    elif typ(g) == t_INT:
        INT_to_mpz(mpq_numref(value), g)
        mpz_set_ui(mpq_denref(value), 1)
    else:
        pari_err(e_TYPE, <char*>"conversion to mpq", g)
